import os

import astropy.units as u
import numpy as np
from numpy.testing import assert_allclose
import pytest
from specutils import Spectrum

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS
from jdaviz.core.marks import MarkersMark, DistanceLine
from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_NoWCS
from bqplot.scales import LinearScale


def _get_markers_from_viewer(viewer):
    return [m for m in viewer.figure.marks if isinstance(m, MarkersMark)][0]


def _assert_dict_allclose(dict1, dict2):
    assert dict1.keys() == dict2.keys()
    for k, v in dict1.items():
        if isinstance(v, float):
            assert_allclose(v, dict2.get(k))
        elif isinstance(v, (tuple, list)):
            assert_allclose(np.asarray(v), np.asarray(dict2.get(k)))
        else:
            assert v == dict2.get(k)


def test_markers_cubeviz(tmp_path, cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    sb_unit = 'Jy / pix2'  # cubes loaded in Jy have sb unit of Jy / pix2
    flux_unit = 'Jy'

    label_mouseover = cubeviz_helper._coords_info

    mp = cubeviz_helper.plugins['Markers']
    mp.keep_active = True
    exp = cubeviz_helper.plugins['Export']

    # no marks yet, so table does not yet appear in export plugin
    assert "Markers: table" not in exp.plugin_table.choices

    # test event in flux viewer
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})

    assert label_mouseover.as_text() == (f'Pixel x=00.0 y=00.0 Value +8.00000e+00 {sb_unit}',
                                         'World 13h39m59.9731s +27d00m00.3600s (ICRS)',
                                         '204.9998877673 27.0001000000 (deg)')

    _assert_dict_allclose(label_mouseover.as_dict(), {'axes_x': 0,
                                                      'axes_x:unit': 'pix',
                                                      'axes_y': 0,
                                                      'axes_y:unit': 'pix',
                                                      'data_label': 'test[FLUX]',
                                                      'slice': 1.0,
                                                      'spectral_axis': 4.62360027696835e-07,
                                                      'spectral_axis:unit': 'm',
                                                      'world_ra': 204.99988776727642,
                                                      'world_dec': 27.000099999955538,
                                                      'world:unreliable': False,
                                                      'pixel_x': 0,
                                                      'pixel_y': 0,
                                                      'pixel:unreliable': False,
                                                      'value': 8.0,
                                                      'value:unit': sb_unit,
                                                      'value:unreliable': False})

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 1
    assert len(_get_markers_from_viewer(fv).x) == 1

    # test event in spectrum viewer (with auto layer detection)
    # x = [4.62280007e-07, 4.62360028e-07]
    # y = [28, 92] Jy / pix2
    label_mouseover._viewer_mouse_event(sv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 4.623e-7, 'y': 0}})

    assert label_mouseover.as_text() == (f'Cursor 4.62300e-07, 0.00000e+00 Value +8.00000e+00 {sb_unit}',  # noqa
                                         'Wave 4.62280e-07 m (0 pix)',
                                         f'Flux 2.80000e+01 {flux_unit}')
    assert label_mouseover.as_dict() == {'data_label': 'Spectrum (sum)',
                                         'axes_x': 4.622800069238093e-07,
                                         'axes_x:unit': 'm',
                                         'slice': 0.0,
                                         'spectral_axis': 4.622800069238093e-07,
                                         'spectral_axis:unit': 'm',
                                         'axes_y': 28.0,
                                         'axes_y:unit': flux_unit,
                                         'value': 28.0,
                                         'value:unit': flux_unit}

    mp._obj._on_viewer_key_event(sv, {'event': 'keydown',
                                      'key': 'm'})
    assert len(mp.export_table()) == 2
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 1

    # test event in spectrum viewer (with cursor only)
    label_mouseover.dataset_selected = 'none'
    label_mouseover._viewer_mouse_event(sv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 4.623e-7, 'y': 0}})

    assert label_mouseover.as_text() == (f'Cursor 4.62300e-07, 0.00000e+00 Value +8.00000e+00 {sb_unit}',  # noqa
                                         '', '')
    assert label_mouseover.as_dict() == {'axes_x': 4.623e-07,
                                         'axes_x:unit': 'm',
                                         'axes_y': 0,
                                         'axes_y:unit': flux_unit,
                                         'data_label': '',
                                         'spectral_axis': 4.623e-07,
                                         'spectral_axis:unit': 'm',
                                         'value': 0,
                                         'value:unit': flux_unit}

    mp._obj._on_viewer_key_event(sv, {'event': 'keydown',
                                      'key': 'm'})

    # test that markers update on unit conversion
    uc = cubeviz_helper.plugins['Unit Conversion']
    uc.flux_unit.selected = 'MJy'

    # find the index of the marker's x-coordinate in the spectral axis of the original input data
    spec = cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True)
    marker_index = np.where(spec.spectral_axis.value == mp._obj.marks['cubeviz-2'].x)
    # use the index to find the associated flux value of the original input
    flux_value = spec.flux[marker_index].value

    # compare the marker's (y) flux value with the flux value of the original input data
    assert_allclose(mp._obj.marks['cubeviz-2'].y[0], flux_value)

    # now check if marks update with a unit that requires spectral density equivalency
    uc.flux_unit.selected = 'erg / (Angstrom s cm2)'

    spec = cubeviz_helper.get_data('Spectrum (sum)', use_display_units=True)
    flux_value = spec.flux[marker_index].value

    assert_allclose(mp._obj.marks['cubeviz-2'].y[0], flux_value)

    assert len(mp.export_table()) == 3
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 2

    # markers hide when plugin closed and keep_active = False
    mp.keep_active = False
    assert _get_markers_from_viewer(fv).visible is False
    assert _get_markers_from_viewer(sv).visible is False

    # markers re-appear when plugin re-opened
    mp._obj.plugin_opened = True
    assert _get_markers_from_viewer(fv).visible is True
    assert _get_markers_from_viewer(sv).visible is True
    assert len(_get_markers_from_viewer(fv).x) == 1
    assert len(_get_markers_from_viewer(sv).x) == 2

    # appears as option in export plugin and exports successfully
    assert "Markers: table" in exp.plugin_table.choices
    filename = str(tmp_path / "cubeviz_export.ecsv")
    exp.filename.auto = False
    exp.filename.value = filename
    exp.plugin_table = "Markers: table"
    exp.export()
    assert os.path.isfile(filename)

    # Also exports to CSV
    filename_2 = str(tmp_path / "cubeviz_export.csv")
    exp.plugin_table_format.selected = 'csv'
    exp.filename.value = filename_2
    exp.export()
    assert os.path.isfile(filename_2)

    # clearing table clears markers
    mp.clear_table()
    assert mp.export_table() is None
    assert len(_get_markers_from_viewer(fv).x) == 0
    assert len(_get_markers_from_viewer(sv).x) == 0


@pytest.mark.parametrize("flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
@pytest.mark.parametrize("new_flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])
def test_markers_cubeviz_flux_unit_conversion(cubeviz_helper,
                                              spectrum1d_cube_custom_fluxunit,
                                              flux_unit, angle_unit, new_flux_unit):
    """
    Test the markers plugin with all possible unit conversions for
    cubes in spectral/photon surface brightness units (e.g. Jy/sr, Jy/pix2).

    The markers plugin should respect the choice of flux and angle
    unit selected in the Unit Conversion plugin, and inputs and results should
    be converted based on selection. All conversions between units in the
    flux dropdown menu in the unit conversion plugin should be supported.
    """

    if new_flux_unit == flux_unit:  # skip 'converting' to same unit
        return

    new_flux_unit_str = new_flux_unit.to_string()

    # load cube with specified unit
    cube = spectrum1d_cube_custom_fluxunit(fluxunit=flux_unit / angle_unit,
                                           shape=(5, 5, 4),
                                           with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="test")

    # get plugins
    mp = cubeviz_helper.plugins['Markers']
    uc = cubeviz_helper.plugins['Unit Conversion']._obj

    mp.keep_active = True

    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    label_mouseover = cubeviz_helper._coords_info
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})

    # set to new unit
    uc.flux_unit.selected = new_flux_unit_str
    new_cube_unit_str = (new_flux_unit / angle_unit).to_string()

    # add a new marker at the same location
    label_mouseover._viewer_mouse_event(fv,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0,
                                                    'y': 0}})
    # mouseover should have changed to new unit
    assert label_mouseover.as_dict()['value:unit'] == new_cube_unit_str

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown',
                                      'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == new_cube_unit_str


def test_markers_specviz2d_unit_conversion(specviz2d_helper):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.AA)
    specviz2d_helper.load_data(spectrum2d)

    uc = specviz2d_helper.plugins['Unit Conversion']
    uc.open_in_tray()
    mp = specviz2d_helper.plugins['Markers']
    mp.keep_active = True

    label_mouseover = specviz2d_helper._coords_info
    viewer2d = specviz2d_helper.app.get_viewer('spectrum-2d-viewer')
    label_mouseover._viewer_mouse_event(viewer2d, {"event": "mousemove",
                                                   "domain": {"x": 6, "y": 3}})
    assert label_mouseover.as_text() == ('Pixel x=06.0 y=03.0 Value +6.00000e+00 MJy',
                                         'Wave 6.00000e+00 Angstrom',
                                         '')
    mp._obj._on_viewer_key_event(viewer2d, {'event': 'keydown',
                                            'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == uc.flux_unit
    assert last_row['spectral_axis:unit'] == uc.spectral_unit

    # ensure marks work with flux conversion where spectral axis is required and
    # spectral axis conversion
    uc.flux_unit = 'erg / (Angstrom s cm2)'
    uc.spectral_unit = 'Ry'
    label_mouseover._viewer_mouse_event(viewer2d, {"event": "mousemove",
                                                   "domain": {"x": 4, "y": 3}})
    assert label_mouseover.as_text() == ('Pixel x=04.0 y=03.0 Value +7.49481e+00 erg / (Angstrom s cm2)',  # noqa
                                         'Wave 2.27817e+02 Ry',
                                         '')
    mp._obj._on_viewer_key_event(viewer2d, {'event': 'keydown',
                                            'key': 'm'})

    # make sure last marker added to table reflects new unit selection
    last_row = mp.export_table()[-1]
    assert last_row['value:unit'] == uc.flux_unit
    assert last_row['spectral_axis:unit'] == uc.spectral_unit

    second_marker_flux_unit = uc.flux_unit.selected
    second_marker_spectral_unit = uc.spectral_unit.selected

    # test edge case two non-native spectral axis required conversions
    uc.flux_unit = 'ph / (Angstrom s cm2)'
    uc.spectral_unit = 'eV'

    # make sure table flux and spectral unit doesn't update
    assert last_row['value:unit'] == second_marker_flux_unit
    assert last_row['spectral_axis:unit'] == second_marker_spectral_unit


class TestImvizMultiLayer(BaseImviz_WCS_NoWCS):
    def test_markers_layer_cycle(self):
        label_mouseover = self.imviz._coords_info

        mp = self.imviz.plugins['Markers']
        mp._obj.plugin_opened = True

        # cycle through dataset options (used for both coords info and markers)
        assert label_mouseover.dataset.choices == ['auto', 'none',
                                                   'has_wcs[SCI,1]',
                                                   'no_wcs[SCI,1]']
        assert label_mouseover.dataset.selected == 'auto'

        # auto (top, no WCS) layer
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0 Value +5.50000e+01', '', '')
        assert label_mouseover.as_dict() == {'axes_x': 5,
                                             'axes_x:unit': 'pix',
                                             'axes_y': 5,
                                             'axes_y:unit': 'pix',
                                             'data_label': 'no_wcs[SCI,1]',
                                             'pixel_x': 5.0,
                                             'pixel_y': 5.0,
                                             'pixel:unreliable': False,
                                             'value': 55.0,
                                             'value:unit': '',
                                             'value:unreliable': False}

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 1

        # no layer (cursor position only)
        label_mouseover.dataset.select_next()
        assert label_mouseover.dataset.selected == 'none'

        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0', '', '')
        assert label_mouseover.as_dict() == {'axes_x': 5,
                                             'axes_x:unit': 'pix',
                                             'axes_y': 5,
                                             'axes_y:unit': 'pix',
                                             'data_label': '',
                                             'pixel_x': 5.0,
                                             'pixel_y': 5.0,
                                             'pixel:unreliable': False}

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 2

        # non-default layer (with WCS)
        label_mouseover.dataset.selected = 'has_wcs[SCI,1]'
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})

        assert label_mouseover.as_text() == ('Pixel x=05.0 y=05.0 Value +5.50000e+01',
                                             'World 22h30m04.5107s -20d49m54.9990s (ICRS)',
                                             '337.5187947654 -20.8319441647 (deg)')
        _assert_dict_allclose(label_mouseover.as_dict(), {'axes_x': 5,
                                                          'axes_x:unit': 'pix',
                                                          'axes_y': 5,
                                                          'axes_y:unit': 'pix',
                                                          'data_label': 'has_wcs[SCI,1]',
                                                          'world_ra': 337.5187947653852,
                                                          'world_dec': -20.831944164705973,
                                                          'world:unreliable': False,
                                                          'pixel_x': 5.0,
                                                          'pixel_y': 5.0,
                                                          'pixel:unreliable': False,
                                                          'value': 55.0,
                                                          'value:unit': '',
                                                          'value:unreliable': False})

        mp._obj._on_viewer_key_event(self.viewer, {'event': 'keydown',
                                                   'key': 'm'})
        assert len(_get_markers_from_viewer(self.viewer).x) == 3

    def test_markers_custom_viewer(self):
        label_mouseover = self.imviz._coords_info

        mp = self.imviz.plugins['Markers']
        mp._obj.plugin_opened = True

        nv = self.imviz.create_image_viewer()
        self.imviz.app.add_data_to_viewer('imviz-1', 'has_wcs[SCI,1]')

        assert label_mouseover.dataset.choices == ['auto', 'none',
                                                   'has_wcs[SCI,1]',
                                                   'no_wcs[SCI,1]']
        assert label_mouseover.dataset.selected == 'auto'

        # top-layer in default viewer is no_wcs[SCI,1]
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'no_wcs[SCI,1]'

        # top-layer in new viewer is has_wcs[SCI,1]
        label_mouseover._viewer_mouse_event(nv,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'has_wcs[SCI,1]'

        # choosing a dataset that is in one viewer but not the other
        label_mouseover.dataset.selected = 'no_wcs[SCI,1]'
        label_mouseover._viewer_mouse_event(self.viewer,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == 'no_wcs[SCI,1]'

        label_mouseover._viewer_mouse_event(nv,
                                            {'event': 'mousemove',
                                             'domain': {'x': 5, 'y': 5}})
        assert label_mouseover.as_dict()['data_label'] == ''


def _get_distance_marks_from_viewer(viewer):
    """Helper function to retrieve DistanceLine marks from a viewer's figure."""
    return [m for m in viewer.figure.marks if isinstance(m, DistanceLine)]


def test_distance_tool_live_preview_image(cubeviz_helper, spectrum1d_cube):
    """Tests the rubber band line and live UI update in an image viewer."""
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 1, 'y': 2}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd'})
    assert mp._obj.distance_display == "..."

    mp._obj._on_mouse_move_while_drawing({'domain': {'x': 4, 'y': 6}})
    # In a WCS-aware viewer, distance is in sky units, not pixels.
    # We check for the correct unit instead of a brittle value check.
    assert mp._obj.distance_display.endswith("arcsec")

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd'})
    assert len(mp._obj.measurements_table.items) == 1
    # The final table entry should also be in sky units.
    assert mp._obj.measurements_table.items[0]['Separation (arcsec)'] != "N/A"


def test_distance_tool_profile(cubeviz_helper, spectrum1d_cube):
    """Tests the rubber band line and live UI update in a profile viewer."""
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    # Set the starting point and get the actual data coordinates from coords_info
    label_mouseover._viewer_mouse_event(sv, {'event': 'mousemove', 'domain': {'x': 4.623e-7,
                                                                              'y': 20}})
    p1_coords = label_mouseover.as_dict()
    mp._obj._on_viewer_key_event(sv, {'event': 'keydown', 'key': 'd'})

    # Set the end point and get the actual data coordinates from coords_info
    x2_cursor, y2_cursor = 4.624e-7, 80
    label_mouseover._viewer_mouse_event(sv, {'event': 'mousemove', 'domain': {'x': x2_cursor,
                                                                              'y': y2_cursor}})
    p2_coords = label_mouseover.as_dict()
    mp._obj._on_mouse_move_while_drawing({'domain': {'x': x2_cursor, 'y': y2_cursor}})

    # Calculate expected deltas based on what the coords_info tool reports
    dx = p2_coords['axes_x'] - p1_coords['axes_x']
    dy = p2_coords['axes_y'] - p1_coords['axes_y']
    assert mp._obj.distance_display == f"Δx: {dx:.4g} m, Δy: {dy:.4g} Jy"

    # Finish measurement and check the final table entry
    mp._obj._on_viewer_key_event(sv, {'event': 'keydown', 'key': 'd'})
    assert len(mp._obj.measurements_table.items) == 1
    profile_row = mp._obj.measurements_table.items[0]

    # Assert that the new Start/End columns are populated correctly
    assert profile_row['Start X'] == f"{p1_coords['axes_x']:.4g} m"
    assert profile_row['End X'] == f"{p2_coords['axes_x']:.4g} m"
    assert profile_row['Start Y'] == f"{p1_coords['axes_y']:.4g} Jy"
    assert profile_row['End Y'] == f"{p2_coords['axes_y']:.4g} Jy"
    # Assert that the delta columns are still correct
    assert profile_row['Δx'] == f"{dx:.4g} m"
    assert profile_row['Δy'] == f"{dy:.4g} Jy"

    # Test distance mark location update after unit conversion
    marks = _get_distance_marks_from_viewer(sv)
    assert_allclose(marks[0].y, [28, 80])
    uc = cubeviz_helper.plugins['Unit Conversion']
    uc.flux_unit.selected = 'MJy'
    assert_allclose(marks[0].y, [2.8e-05, 8.0e-05])


def test_distance_tool_snapping(cubeviz_helper, spectrum1d_cube):
    """Tests that the distance tool can snap to existing markers."""
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 0, 'y': 0}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'm'})

    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 0.1, 'y': 0.1}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd', 'altKey': True})
    assert_allclose(mp._obj._distance_first_point['axes_x'], 0)

    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd', 'altKey': True})
    # Must access the table via the internal ._obj
    assert len(mp._obj.measurements_table.items) == 1
    assert mp._obj.measurements_table.items[0]['Distance (pix)'] == "0.00"


def test_distance_tool_clearing(cubeviz_helper, spectrum1d_cube):
    """Tests that clearing the measurements table removes marks from the viewer."""
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    # 1. Create a measurement.
    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 1, 'y': 2}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd'})
    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 4, 'y': 6}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd'})
    assert len(mp._obj.measurements_table.items) == 1

    # 2. Call the public clear_table() method, which is what the UI button does.
    mp._obj.measurements_table.clear_table()

    # 3. Assert that the table is empty and viewer marks are gone.
    assert len(mp._obj.measurements_table.items) == 0
    assert len(_get_distance_marks_from_viewer(fv)) == 0


def test_distance_tool_reset_on_orientation_change(cubeviz_helper, spectrum1d_cube):
    """Tests that an in-progress measurement is cancelled on orientation change."""
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    label_mouseover._viewer_mouse_event(fv, {'event': 'mousemove', 'domain': {'x': 1, 'y': 2}})
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd'})
    assert mp._obj._distance_first_point is not None

    mp._obj._on_alignment_change()
    assert mp._obj._distance_first_point is None
    assert mp._obj.distance_display == "N/A"


def test_distance_tool_imviz_pixel_only(imviz_helper):
    """Tests the distance tool in a pixel-only context in Imviz."""
    # Use the imviz_helper fixture and load a simple array with no WCS
    imviz_helper.load_data(np.zeros((20, 20)))
    iv = imviz_helper.app.get_viewer('imviz-0')
    mp = imviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = imviz_helper._coords_info

    # In Imviz, we can force pixel-based alignment
    imviz_helper.app._align_by = 'pixels'

    # Start measurement
    label_mouseover._viewer_mouse_event(iv, {'event': 'mousemove', 'domain': {'x': 2, 'y': 3}})
    mp._obj._on_viewer_key_event(iv, {'event': 'keydown', 'key': 'd'})

    # Update the coords_info plugin by simulating the mouse moving to the new location.
    label_mouseover._viewer_mouse_event(iv, {'event': 'mousemove', 'domain': {'x': 5, 'y': 7}})

    # Now, check the distance display after the move.
    mp._obj._on_mouse_move_while_drawing({'domain': {'x': 5, 'y': 7}})
    # Distance = sqrt((5-2)^2 + (7-3)^2) = sqrt(9 + 16) = 5
    assert mp._obj.distance_display.startswith("5.00 pix")

    # Finish measurement
    mp._obj._on_viewer_key_event(iv, {'event': 'keydown', 'key': 'd'})
    assert len(mp._obj.measurements_table.items) == 1
    assert mp._obj.measurements_table.items[0]['Distance (pix)'].startswith("5.00")


def test_markers_specviz_config(specviz_helper, spectrum1d):
    """Tests the marker plugin's basic functionality in a Specviz context."""
    specviz_helper.load_data(spectrum1d)
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    mp = specviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = specviz_helper._coords_info

    # Check that the table headers are correct for Specviz (no RA/Dec)
    assert 'world_ra' not in mp._obj.table.headers_avail
    assert 'index' in mp._obj.table.headers_avail

    # Place a marker
    label_mouseover._viewer_mouse_event(sv, {'event': 'mousemove', 'domain': {'x': 7000, 'y': 1}})
    mp._obj._on_viewer_key_event(sv, {'event': 'keydown', 'key': 'm'})

    # Assert that the marker was added to the table
    assert len(mp.export_table()) == 1
    assert mp.export_table()['spectral_axis'][0] > 0


def test_distance_tool_snapping_by_pixel_scale(cubeviz_helper, spectrum1d_cube):
    """
    Tests the snapping functionality that relies on viewer pixel scales,
    covering both image and profile viewer cases to improve test coverage.
    """
    cubeviz_helper.load_data(spectrum1d_cube, "test")
    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    sv = cubeviz_helper.app.get_viewer('spectrum-viewer')
    mp = cubeviz_helper.plugins['Markers']
    mp.open_in_tray()
    label_mouseover = cubeviz_helper._coords_info

    # --- Test 1: Image Viewer (2D distance snapping) ---
    # Manually attach real bqplot Scale objects with the necessary attributes
    # to ensure the primary snapping logic is triggered.
    mock_x_scale_img = LinearScale()
    mock_x_scale_img.domain = (0, 10)
    mock_x_scale_img.range = (0, 100)
    mock_y_scale_img = LinearScale()
    mock_y_scale_img.domain = (0, 10)
    mock_y_scale_img.range = (0, 100)
    fv.scales = {'x': mock_x_scale_img, 'y': mock_y_scale_img}

    marker_x_img, marker_y_img = 2, 3
    label_mouseover._viewer_mouse_event(
        fv, {'event': 'mousemove', 'domain': {'x': marker_x_img, 'y': marker_y_img}}
    )
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'm'})
    label_mouseover._viewer_mouse_event(
        fv, {'event': 'mousemove', 'domain': {'x': marker_x_img + 0.1, 'y': marker_y_img + 0.1}}
    )
    mp._obj._on_viewer_key_event(fv, {'event': 'keydown', 'key': 'd', 'altKey': True})

    assert mp._obj._distance_first_point is not None
    assert_allclose(mp._obj._distance_first_point['axes_x'], marker_x_img)
    assert_allclose(mp._obj._distance_first_point['axes_y'], marker_y_img)

    mp.clear_table()
    mp._obj._on_alignment_change()

    # --- Test 2: Profile Viewer (1D distance snapping) ---
    # Manually attach scale properties to the spectrum viewer.
    mock_x_scale_spec = LinearScale()
    mock_x_scale_spec.domain = (4.6e-7, 4.7e-7)
    mock_x_scale_spec.range = (0, 500)
    mock_y_scale_spec = LinearScale()
    mock_y_scale_spec.domain = (0, 100)
    mock_y_scale_spec.range = (0, 300)
    sv.scales = {'x': mock_x_scale_spec, 'y': mock_y_scale_spec}

    marker_x_spec = 4.6236e-07
    marker_y_spec = 50
    label_mouseover._viewer_mouse_event(
        sv, {'event': 'mousemove', 'domain': {'x': marker_x_spec, 'y': marker_y_spec}}
    )
    marker_coords = label_mouseover.as_dict()
    mp._obj._on_viewer_key_event(sv, {'event': 'keydown', 'key': 'm'})
    label_mouseover._viewer_mouse_event(
        sv, {'event': 'mousemove', 'domain': {'x': marker_x_spec + 1e-9, 'y': marker_y_spec + 20}}
    )
    mp._obj._on_viewer_key_event(sv, {'event': 'keydown', 'key': 'd', 'altKey': True})

    assert mp._obj._distance_first_point is not None
    assert_allclose(mp._obj._distance_first_point['axes_x'], marker_coords['axes_x'])



from functools import cached_property
import asdf
import warnings

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry

try:
    from roman_datamodels import datamodels as rdd
except ImportError:
    HAS_ROMAN_DATAMODELS = False
else:
    HAS_ROMAN_DATAMODELS = True

__all__ = ['ASDFParser']


@loader_parser_registry('asdf')
class ASDFParser(BaseParser):

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'imviz'):
            # NOTE: temporary during deconfig process
            return False

        try:
            self.output
        except Exception:
            return False

        return True

    @cached_property
    def output(self):
        if HAS_ROMAN_DATAMODELS:
            return rdd.open(self.input)
        else:
            warnings.warn(
                f"{self.input} should be opened with the `roman_datamodels` package, "
                "which is not installed in this environment. To install optional "
                "jdaviz dependencies for Roman, you can run:  \n\n"
                "pip install -U jdaviz[roman]\n\n"
                "This file will be loaded with the `asdf` package instead.\n\n",
                UserWarning
            )
        return asdf.open(self.input)



from .parser import *  # noqa
from .fits import *  # noqa
from .jpg_png import *  # noqa
from .asdf import *  # noqa
from .regions import *  # noqa
from .specutils import *  # noqa
from .object import *  # noqa


from traitlets import Any, Int
from solara import FileDropMultiple
from ipywidgets import widget_serialization
import io
import os
import reacton

from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi


@loader_resolver_registry('file drop')
class FileDropResolver(BaseResolver):
    template_file = __file__, "file_drop.vue"

    progress = Int(100).tag(sync=True)
    nfiles = Int(0).tag(sync=True)
    file_drop_widget = Any().tag(sync=True, **widget_serialization)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._file_info = None

        self.file_drop_widget_el = FileDropMultiple(label="Drop file here",
                                                    on_total_progress=self._on_total_progress,
                                                    on_file=self._on_file_updated,
                                                    lazy=False)
        self.file_drop_widget, rc = reacton.render(self.file_drop_widget_el)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=[])

    @property
    def is_valid(self):
        return True

    @property
    def default_label(self):
        # Use the first file name as the default label, if available.
        # Otherwise, return None.
        if self._file_info and 'name' in self._file_info:
            return os.path.splitext(self._file_info['name'])[0]
        return None

    def _on_total_progress(self, progress):
        self.progress = progress

    def _on_file_updated(self, file_infos):
        self.nfiles = len(file_infos)
        self._file_info = file_infos[0]
        self._update_format_items()
        self.progress = 100

    def __call__(self, local_path=None, timeout=60):
        # this will return a bytes object of the file contents
        return io.BytesIO(self._file_info.get('data'))



import numpy as np
from functools import cached_property

from traitlets import Bool, List, Unicode, observe

from jdaviz.configs.mosviz.plugins.viewers import Spectrum1DViewer
from jdaviz.core.events import (SnackbarMessage, NewViewerMessage,
                                ViewerVisibleLayersChangedMessage)
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        SelectPluginComponent,
                                        DatasetSelect,
                                        AddResults,
                                        skip_if_no_updates_since_last_active,
                                        skip_if_not_tray_instance,
                                        skip_if_not_relevant,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty
from jdaviz.core.marks import PluginMarkCollection, PluginLine

from astropy.modeling import models
from astropy.nddata import StdDevUncertainty, VarianceUncertainty, UnknownUncertainty
from specutils import Spectrum
from specreduce import tracing
from specreduce import background
from specreduce import extract

__all__ = ['SpectralExtraction2D']

_model_cls = {'Spline': models.Spline1D,
              'Polynomial': models.Polynomial1D,
              'Legendre': models.Legendre1D,
              'Chebyshev': models.Chebyshev1D}


@tray_registry('spectral-extraction-2d', label="2D Spectral Extraction",
               category="data:reduction")
class SpectralExtraction2D(PluginTemplateMixin):
    """
    The Spectral Extraction 2D plugin exposes specreduce methods for tracing,
    background subtraction, and spectral extraction from 2D spectra.
    See the :ref:`2D Spectral Extraction Plugin Documentation <specviz2d-spectral-extraction>`
    for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :attr:`interactive_extract`
    * ``trace_dataset`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the input dataset for generating the trace.
    * ``trace_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the type of trace to be generated.
    * ``trace_peak_method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      only applicable if ``trace_type`` is not ``Flat``.
    * :attr:`trace_pixel` :
      pixel of the trace.  If ``trace_type`` is not ``Flat``, then this
      is the "guess" for the automated trace.
    * :attr:`trace_do_binning` :
      only applicable if ``trace_type`` is not ``Flat``.  Bin the input data when fitting the
      trace.
    * :attr:`trace_bins` :
      only applicable if ``trace_type`` is not ``Flat`` and ``trace_do_binning``.
    * :attr:`trace_window` :
      full width of the trace.
    * :meth:`import_trace`
    * :meth:`export_trace`
    * ``bg_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      controls the input dataset for generating the background.
    * ``bg_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      controls the type of background to be generated.
    * :attr:`bg_trace_pixel` :
      only applicable if ``bg_type`` is set to ``Manual``
    * :attr:`bg_separation` :
      only applicable if ``bg_type`` set set to ``OneSided`` or ``TwoSided``.
      Separation from the referenced trace for the center of each of the background window(s).
    * :attr:`bg_width` :
      full width of each background window(s).
    * ``bg_statistic`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      statistic to use when computing the background.  'Average' will account for partial pixel
      weights, 'Median' will include all partial pixels.
    * ``bg_add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * ``bg_sub_add_results``
    * :meth:`import_bg`
    * :meth:`export_bg`
    * :meth:`export_bg_spectrum`
    * :meth:`export_bg_img`
    * :meth:`export_bg_sub`
    * ``ext_dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      controls the input dataset for generating the extracted spectrum.  "From Plugin" will
      use the background-subtracted image as defined by the background options above.  To skip
      background extraction, select the original input 2D spectrum.
    * ``ext_trace`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`)
    * ``ext_type`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`)
    * :attr:`ext_width` :
      full width of the extraction window.
    * ``horne_ext_profile`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      For Horne extract, choice of 'Gaussian' or 'Self (interpolated)' to use
      empirical profile from data.
    * :attr:`self_prof_n_bins` :
      Number of bins to use when computing the self-derived profile for Horne Extract.
    * :attr:`self_prof_interp_degree_x` :
      Interpolation degree (in X) to use when computing the self-derived profile
      for Horne Extract.
    * :attr:`self_prof_interp_degree_y` :
      Interpolation degree (in Y) to use when computing the self-derived profile
      for Horne Extract.
    * ``ext_add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`import_extract`
    * :meth:`export_extract`
    * :meth:`export_extract_spectrum`
    """
    dialog = Bool(False).tag(sync=True)
    template_file = __file__, "spectral_extraction.vue"
    uses_active_status = Bool(True).tag(sync=True)

    active_step = Unicode().tag(sync=True)

    # SETTINGS
    interactive_extract = Bool(True).tag(sync=True)

    # TRACE
    trace_trace_items = List().tag(sync=True)
    trace_trace_selected = Unicode().tag(sync=True)

    trace_offset = IntHandleEmpty(0).tag(sync=True)

    trace_dataset_items = List().tag(sync=True)
    trace_dataset_selected = Unicode().tag(sync=True)

    trace_type_items = List().tag(sync=True)
    trace_type_selected = Unicode().tag(sync=True)

    trace_pixel = FloatHandleEmpty(0).tag(sync=True)
    trace_order = IntHandleEmpty(3).tag(sync=True)

    trace_peak_method_items = List().tag(sync=True)
    trace_peak_method_selected = Unicode().tag(sync=True)

    trace_do_binning = Bool(True).tag(sync=True)
    trace_bins = IntHandleEmpty(20).tag(sync=True)
    trace_window = IntHandleEmpty(0).tag(sync=True)

    trace_results_label = Unicode().tag(sync=True)
    trace_results_label_default = Unicode().tag(sync=True)
    trace_results_label_auto = Bool(True).tag(sync=True)
    trace_results_label_invalid_msg = Unicode('').tag(sync=True)
    trace_results_label_overwrite = Bool().tag(sync=True)
    trace_add_to_viewer_items = List().tag(sync=True)
    trace_add_to_viewer_selected = Unicode().tag(sync=True)
    trace_spinner = Bool(False).tag(sync=True)

    # BACKGROUND
    bg_dataset_items = List().tag(sync=True)
    bg_dataset_selected = Unicode().tag(sync=True)

    bg_type_items = List().tag(sync=True)
    bg_type_selected = Unicode().tag(sync=True)

    bg_trace_items = List().tag(sync=True)
    bg_trace_selected = Unicode().tag(sync=True)

    bg_trace_pixel = FloatHandleEmpty(0).tag(sync=True)

    bg_statistic_items = List().tag(sync=True)
    bg_statistic_selected = Unicode().tag(sync=True)

    bg_separation = FloatHandleEmpty(0).tag(sync=True)
    bg_width = FloatHandleEmpty(0).tag(sync=True)

    bg_results_label = Unicode().tag(sync=True)
    bg_results_label_default = Unicode().tag(sync=True)
    bg_results_label_auto = Bool(True).tag(sync=True)
    bg_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_results_label_overwrite = Bool().tag(sync=True)
    bg_add_to_viewer_items = List().tag(sync=True)
    bg_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_img_spinner = Bool(False).tag(sync=True)

    bg_spec_results_label = Unicode().tag(sync=True)
    bg_spec_results_label_default = Unicode().tag(sync=True)
    bg_spec_results_label_auto = Bool(True).tag(sync=True)
    bg_spec_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_spec_results_label_overwrite = Bool().tag(sync=True)
    bg_spec_add_to_viewer_items = List().tag(sync=True)
    bg_spec_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_spec_spinner = Bool(False).tag(sync=True)

    bg_sub_results_label = Unicode().tag(sync=True)
    bg_sub_results_label_default = Unicode().tag(sync=True)
    bg_sub_results_label_auto = Bool(True).tag(sync=True)
    bg_sub_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_sub_results_label_overwrite = Bool().tag(sync=True)
    bg_sub_add_to_viewer_items = List().tag(sync=True)
    bg_sub_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_sub_spinner = Bool(False).tag(sync=True)

    # EXTRACT
    ext_dataset_items = List().tag(sync=True)
    ext_dataset_selected = Unicode().tag(sync=True)

    ext_trace_items = List().tag(sync=True)
    ext_trace_selected = Unicode().tag(sync=True)

    ext_type_items = List().tag(sync=True)
    ext_type_selected = Unicode().tag(sync=True)

    horne_ext_profile_items = List().tag(sync=True)
    horne_ext_profile_selected = Unicode().tag(sync=True)

    self_prof_n_bins = IntHandleEmpty(10).tag(sync=True)
    self_prof_interp_degree_x = IntHandleEmpty(1).tag(sync=True)
    self_prof_interp_degree_y = IntHandleEmpty(1).tag(sync=True)

    ext_width = FloatHandleEmpty(0).tag(sync=True)

    ext_uncert_warn = Bool(False).tag(sync=True)
    ext_specreduce_err = Unicode().tag(sync=True)

    ext_results_label = Unicode().tag(sync=True)
    ext_results_label_default = Unicode().tag(sync=True)
    ext_results_label_auto = Bool(True).tag(sync=True)
    ext_results_label_invalid_msg = Unicode('').tag(sync=True)
    ext_results_label_overwrite = Bool().tag(sync=True)
    ext_add_to_viewer_items = List().tag(sync=True)
    ext_add_to_viewer_selected = Unicode().tag(sync=True)
    # uses default "spinner"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Extract 1D spectrum from 2D image.'

        # TRACE
        self.trace_trace = DatasetSelect(self,
                                         'trace_trace_items',
                                         'trace_trace_selected',
                                         default_text='New Trace',
                                         filters=['is_trace'])

        self.trace_dataset = DatasetSelect(self,
                                           'trace_dataset_items',
                                           'trace_dataset_selected',
                                           filters=['layer_in_spectrum_2d_viewer', 'not_trace'])
        self.trace_dataset.get_data_cls = Spectrum

        self.trace_type = SelectPluginComponent(self,
                                                items='trace_type_items',
                                                selected='trace_type_selected',
                                                manual_options=['Flat', 'Polynomial',
                                                                'Legendre', 'Chebyshev',
                                                                'Spline'])

        self.trace_peak_method = SelectPluginComponent(self,
                                                       items='trace_peak_method_items',
                                                       selected='trace_peak_method_selected',
                                                       manual_options=['Gaussian', 'Centroid', 'Max'])  # noqa

        self.trace_add_results = AddResults(self, 'trace_results_label',
                                            'trace_results_label_default',
                                            'trace_results_label_auto',
                                            'trace_results_label_invalid_msg',
                                            'trace_results_label_overwrite',
                                            'trace_add_to_viewer_items',
                                            'trace_add_to_viewer_selected')
        self.trace_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.trace_results_label_default = 'trace'

        # BACKGROUND
        self.bg_dataset = DatasetSelect(self,
                                        'bg_dataset_items',
                                        'bg_dataset_selected',
                                        filters=['layer_in_spectrum_2d_viewer', 'not_trace'])

        self.bg_type = SelectPluginComponent(self,
                                             items='bg_type_items',
                                             selected='bg_type_selected',
                                             manual_options=['TwoSided', 'OneSided', 'Manual'])

        self.bg_trace = DatasetSelect(self,
                                      items='bg_trace_items',
                                      selected='bg_trace_selected',
                                      default_text='From Plugin',
                                      filters=['is_trace'])

        self.bg_statistic = SelectPluginComponent(self,
                                                  items='bg_statistic_items',
                                                  selected='bg_statistic_selected',
                                                  manual_options=['Average', 'Median'])

        self.bg_add_results = AddResults(self, 'bg_results_label',
                                         'bg_results_label_default',
                                         'bg_results_label_auto',
                                         'bg_results_label_invalid_msg',
                                         'bg_results_label_overwrite',
                                         'bg_add_to_viewer_items',
                                         'bg_add_to_viewer_selected')
        self.bg_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.bg_results_label_default = 'background'

        self.bg_spec_add_results = AddResults(self, 'bg_spec_results_label',
                                              'bg_spec_results_label_default',
                                              'bg_spec_results_label_auto',
                                              'bg_spec_results_label_invalid_msg',
                                              'bg_spec_results_label_overwrite',
                                              'bg_spec_add_to_viewer_items',
                                              'bg_spec_add_to_viewer_selected')
        self.bg_spec_add_results.viewer.filters = ['is_spectrum_viewer']
        self.bg_spec_results_label_default = 'background-spectrum'

        self.bg_sub_add_results = AddResults(self, 'bg_sub_results_label',
                                             'bg_sub_results_label_default',
                                             'bg_sub_results_label_auto',
                                             'bg_sub_results_label_invalid_msg',
                                             'bg_sub_results_label_overwrite',
                                             'bg_sub_add_to_viewer_items',
                                             'bg_sub_add_to_viewer_selected')
        self.bg_sub_add_results.viewer.filters = ['is_spectrum_2d_viewer']
        self.bg_sub_results_label_default = 'background-subtracted'

        # EXTRACT
        self.ext_dataset = DatasetSelect(self,
                                         'ext_dataset_items',
                                         'ext_dataset_selected',
                                         default_text='From Plugin',
                                         filters=['layer_in_spectrum_2d_viewer', 'not_trace'])

        self.ext_trace = DatasetSelect(self,
                                       items='ext_trace_items',
                                       selected='ext_trace_selected',
                                       default_text='From Plugin',
                                       filters=['is_trace'])

        self.ext_type = SelectPluginComponent(self,
                                              items='ext_type_items',
                                              selected='ext_type_selected',
                                              manual_options=['Boxcar', 'Horne'])

        self.horne_ext_profile = SelectPluginComponent(self,
                                                       items='horne_ext_profile_items',
                                                       selected='horne_ext_profile_selected',
                                                       manual_options=['Gaussian', 'Self (interpolated)'])  # noqa

        self.ext_add_results = AddResults(self, 'ext_results_label',
                                          'ext_results_label_default',
                                          'ext_results_label_auto',
                                          'ext_results_label_invalid_msg',
                                          'ext_results_label_overwrite',
                                          'ext_add_to_viewer_items',
                                          'ext_add_to_viewer_selected')
        self.ext_add_results.viewer.filters = ['is_spectrum_viewer']
        # NOTE: defaults to overwriting original spectrum
        self.ext_add_results.label_whitelist_overwrite = ['1D Spectrum', '2D Spectrum (auto-ext)']
        self.ext_results_label_default = '2D Spectrum (auto-ext)'

        self.app.hub.subscribe(self, ViewerVisibleLayersChangedMessage,
                               lambda _: self._update_plugin_marks())

        if self.config == "deconfigged":
            self.observe_traitlets_for_relevancy(traitlets_to_observe=['trace_dataset_items'])

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('interactive_extract',
                                           'trace_dataset', 'trace_type',
                                           'trace_order', 'trace_peak_method',
                                           'trace_pixel',
                                           'trace_do_binning', 'trace_bins', 'trace_window',
                                           'import_trace',
                                           'export_trace',
                                           'bg_dataset', 'bg_type',
                                           'bg_trace_pixel', 'bg_separation', 'bg_width',
                                           'bg_statistic',
                                           'bg_add_results', 'bg_sub_add_results',
                                           'import_bg',
                                           'export_bg', 'export_bg_spectrum',
                                           'export_bg_img', 'export_bg_sub',
                                           'ext_dataset', 'ext_trace', 'ext_type',
                                           'ext_width', 'ext_add_results',
                                           'horne_ext_profile',
                                           'self_prof_n_bins',
                                           'self_prof_interp_degree_x',
                                           'self_prof_interp_degree_y',
                                           'import_extract',
                                           'export_extract', 'export_extract_spectrum'))

    def _clear_default_inputs(self):
        self.trace_pixel = 0
        self.trace_window = 0
        self.bg_trace_pixel = 0
        self.bg_separation = 0
        self.bg_width = 0
        self.ext_width = 0

    @observe('irrelevant_msg')
    def _updates_when_becoming_relevant(self, msg):
        if msg.get('new') != '':
            return
        # reset all defaults for the selected trace dataset, _trace_dataset_selected
        # should be triggered shortly after
        self._clear_default_inputs()

    def _extract_in_new_instance(self, dataset=None, add_data=False):
        # create a new instance of the 2D Spectral Extraction plugin (to not
        # affect the instance in the tray) and extract the entire cube with defaults.
        plg = self.new()
        # all other settings remain at their plugin defaults
        plg._clear_default_inputs()
        plg.trace_dataset.selected = self.trace_dataset.selected if dataset is None else dataset
        plg._trace_dataset_selected()  # should only be necessary if default dataset
        return plg.export_extract_spectrum(add_data=add_data)

    @observe('trace_dataset_selected')
    @skip_if_not_relevant()
    def _trace_dataset_selected(self, msg=None):
        if not hasattr(self, 'trace_dataset'):
            # happens when first initializing plugin outside of tray
            return
        if not len(self.trace_dataset.selected):
            return

        width = self.trace_dataset.get_selected_spectrum(use_display_units=True).shape[0]
        # estimate the pixel number by taking the median of the brightest pixel index
        # in each column, ignoring columns where the sum in that column is not
        # positive (ie. columns of all zeros or nans)
        trace_flux = self.trace_dataset.get_selected_spectrum(use_display_units=True).flux
        trace_flux_ignore_zeros = trace_flux[:, np.nansum(trace_flux, axis=0) != 0]
        if trace_flux_ignore_zeros.shape[1] == 0:
            # default to trace in middle of image
            brightest_pixel = int(trace_flux.shape[0]/2)
        else:
            brightest_pixel = int(np.nanmedian(np.nanargmax(trace_flux_ignore_zeros, axis=0)))
        # do not allow to be an edge pixel
        if brightest_pixel < 1:
            brightest_pixel = 1
        if brightest_pixel > width - 1:
            brightest_pixel = width - 1
        distance_from_edge = min(brightest_pixel, width-brightest_pixel)
        # default width will be 10% of cross-dispersion "height",
        # but no larger than distance from the edge
        default_bg_width = int(np.ceil(width / 10))
        default_width = min(default_bg_width, distance_from_edge * 2)

        # sign for one-sided and single trace-pixel depending on whether the brightest pixel is
        # above or below the middle of the image
        if default_bg_width * 2 >= distance_from_edge:
            sign = 1 if (brightest_pixel < width / 2) else -1
            default_bg_separation = sign * default_bg_width * 2
        else:
            default_bg_separation = default_bg_width * 2

        if self.trace_pixel == 0:
            self.trace_pixel = brightest_pixel
        if self.trace_window == 0:
            self.trace_window = default_width
        if self.bg_trace_pixel == 0:
            self.bg_trace_pixel = brightest_pixel + default_bg_separation
        if self.bg_separation == 0:
            if default_bg_width * 2 >= distance_from_edge:
                self.bg_type_selected = 'OneSided'
            self.bg_separation = default_bg_separation
        if self.bg_width == 0:
            self.bg_width = default_bg_width
        if self.ext_width == 0:
            self.ext_width = default_width

    def update_marks(self, step=None):
        """
        Manually update the live-preview marks for a given step in spectral extraction.  This API
        mimics opening the plugin and interacting with one of the steps.

        Parameters
        ----------
        step : str
            Step in the extraction process to visualize.  Must be one of: 'trace', 'bg', 'ext'.
        """
        if step is not None:
            if step == 'trace':
                self._interaction_in_trace_step()
            elif step == 'bg':
                self._interaction_in_bg_step()
            elif step == 'ext':
                self._interaction_in_ext_step()
            elif step == '':
                return
            else:
                raise ValueError("step must be one of: trace, bg, ext")

    @observe('ext_add_to_viewer_selected', 'ext_results_label_overwrite')
    def _spectrum1d_viewer_changed(self, *args):
        self.update_marks(self.active_step)

    @observe('is_active', 'active_step')
    @skip_if_not_tray_instance()
    def _update_plugin_marks(self, msg={}):
        if self.app._jdaviz_helper is None:
            return

        if not len(self.marks):
            # plugin has never been opened, no need to create marks just to hide them,
            # we'll create marks when the plugin is first opened
            return

        if not (self.is_active):
            for step, mark in self.marks.items():
                mark.clear()
            return

        if self.active_step == '':
            # on load, default to 'extract' (this will then trigger the observe to update the marks)
            self.active_step = 'ext'
            return

        marks_info = self.marks_info(include_mark_obj=False)  # viewer + step info for each mark
        viewers = {'1d': self.marks_viewers1d, '2d': self.marks_viewers2d}
        for step, mark in self.marks.items():
            visible = self.active_step in marks_info[step]['steps']
            mark.set_for_viewers('visible', visible, viewers[marks_info[step]['viewer']])
            mark.clear_if_not_in_viewers(viewers[marks_info[step]['viewer']])

    @property
    def marks_viewers2d(self):
        if self.active_step == 'bg':
            return self.bg_dataset.viewers_with_selected_visible
        return self.trace_dataset.viewers_with_selected_visible

    @property
    def marks_viewers1d(self):
        return self.ext_add_results.results_viewers

    def marks_info(self, include_mark_obj=True):
        """
        A dictionary containing each marker name, the viewer(s) (2d/1d spectral)
        it belongs in, which step (bg, trace, extract) and plotting style kwargs.
        """

        markers = {'trace': {'viewer': '2d', 'steps': ['trace', 'bg', 'ext']},
                   'extract': {'viewer': '1d', 'steps': ['trace', 'bg', 'ext']},
                   'bg1_lower': {'viewer': '2d', 'steps': ['bg']},
                   'bg1_upper': {'viewer': '2d', 'steps': ['bg']},
                   'bg1_center': {'viewer': '2d', 'steps': ['bg'], 'kw': {'line_style': 'dotted'}},
                   'bg2_lower': {'viewer': '2d', 'steps': ['bg']},
                   'bg2_upper': {'viewer': '2d', 'steps': ['bg']},
                   'bg2_center': {'viewer': '2d', 'steps': ['bg'], 'kw': {'line_style': 'dotted'}},
                   'bg_spec': {'viewer': '1d', 'steps': ['bg'], 'kw': {'stroke_width': 1}},
                   'ext_lower': {'viewer': '2d', 'steps': ['ext']},
                   'ext_upper': {'viewer': '2d', 'steps': ['ext']}}

        if include_mark_obj:  # add PluginMarkCollection
            for key, d in markers.items():
                kwargs = {'visible': self.is_active} | d.get('kw', {})
                d['mark'] = PluginMarkCollection(PluginLine, **kwargs)

        return markers

    @cached_property
    def marks(self):
        """
        Access the marks created by this plugin in both the spectrum-viewer
        and spectrum-2d-viewer.
        """
        if not self._tray_instance:
            return {}

        return {k: v['mark'] for k, v in self.marks_info().items()}

    @observe('interactive_extract')
    @skip_if_no_updates_since_last_active()
    @skip_if_not_tray_instance()
    @skip_if_not_relevant()
    def _update_interactive_extract(self, event={}):
        # also called by any of the _interaction_in_*_step
        if self.interactive_extract:
            try:
                sp1d = self.export_extract_spectrum(add_data=False)
            except Exception as e:
                # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
                # NOTE: FitTrace or manual background are often giving a
                # "background regions overlapped" error from specreduce
                self.ext_specreduce_err = repr(e)
                self.marks['extract'].clear()
            else:
                self.ext_specreduce_err = ''
                self.marks['extract'].update_xy(sp1d.spectral_axis.value,
                                                sp1d.flux.value,
                                                viewers=self.marks_viewers1d)
        else:
            self.marks['extract'].clear()

        if self.interactive_extract and self.active_step == 'bg':
            try:
                spec = self.export_bg_spectrum()
            except Exception:
                self.marks['bg_spec'].clear()
            else:
                self.marks['bg_spec'].update_xy(spec.spectral_axis,
                                                spec.flux,
                                                viewers=self.marks_viewers1d)
        else:
            self.marks['bg_spec'].clear()

    @observe('is_active', 'trace_dataset_selected', 'trace_type_selected',
             'trace_trace_selected', 'trace_offset', 'trace_order',
             'trace_pixel', 'trace_peak_method_selected',
             'trace_do_binning', 'trace_bins', 'trace_window', 'active_step')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    @skip_if_not_relevant()
    def _interaction_in_trace_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step != 'trace')
                or not self.is_active):
            return

        try:
            trace = self.export_trace(add_data=False)
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            self.marks['trace'].clear()
        else:
            self.marks['trace'].update_xy(range(len(trace.trace)),
                                          trace.trace,
                                          viewers=self.marks_viewers2d)
            self.marks['trace'].line_style = 'solid'

        self._update_interactive_extract(event)

        self.active_step = 'trace'

    @observe('is_active', 'bg_dataset_selected', 'bg_type_selected',
             'bg_trace_selected', 'bg_trace_pixel',
             'bg_separation', 'bg_width', 'bg_statistic_selected', 'active_step')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    @skip_if_not_relevant()
    def _interaction_in_bg_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step != 'bg')
                or not self.is_active):
            return

        try:
            trace = self._get_bg_trace()
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            for mark in ['trace', 'bg1_center', 'bg1_lower', 'bg1_upper',
                         'bg2_center', 'bg2_lower', 'bg2_upper', 'bg_spec']:
                self.marks[mark].clear()
        else:
            xs = range(len(trace.trace))
            self.marks['trace'].update_xy(xs,
                                          trace.trace,
                                          viewers=self.marks_viewers2d)
            self.marks['trace'].line_style = 'dashed'

            if self.bg_type_selected in ['OneSided', 'TwoSided']:
                self.marks['bg1_center'].update_xy(xs,
                                                   trace.trace+self.bg_separation,
                                                   viewers=self.marks_viewers2d)
                self.marks['bg1_lower'].update_xy(xs,
                                                  trace.trace+self.bg_separation-self.bg_width/2,
                                                  viewers=self.marks_viewers2d)
                self.marks['bg1_upper'].update_xy(xs,
                                                  trace.trace+self.bg_separation+self.bg_width/2,
                                                  viewers=self.marks_viewers2d)
            else:
                self.marks['bg1_center'].clear()
                self.marks['bg1_lower'].update_xy(xs,
                                                  trace.trace-self.bg_width/2,
                                                  viewers=self.marks_viewers2d)
                self.marks['bg1_upper'].update_xy(xs,
                                                  trace.trace+self.bg_width/2,
                                                  viewers=self.marks_viewers2d)

            if self.bg_type_selected == 'TwoSided':
                self.marks['bg2_center'].update_xy(xs,
                                                   trace.trace-self.bg_separation,
                                                   viewers=self.marks_viewers2d)
                self.marks['bg2_lower'].update_xy(xs,
                                                  trace.trace-self.bg_separation-self.bg_width/2,
                                                  viewers=self.marks_viewers2d)
                self.marks['bg2_upper'].update_xy(xs,
                                                  trace.trace-self.bg_separation+self.bg_width/2,
                                                  viewers=self.marks_viewers2d)
            else:
                for mark in ['bg2_center', 'bg2_lower', 'bg2_upper']:
                    self.marks[mark].clear()

        self._update_interactive_extract(event)

        self.active_step = 'bg'

    @observe('is_active', 'ext_dataset_selected', 'ext_trace_selected',
             'ext_type_selected', 'ext_width', 'active_step',
             'horne_ext_profile_selected', 'self_prof_n_bins',
             'self_prof_interp_degree_x', 'self_prof_interp_degree_y')
    @skip_if_not_tray_instance()
    @skip_if_no_updates_since_last_active()
    @skip_if_not_relevant()
    def _interaction_in_ext_step(self, event={}):
        if ((event.get('name', '') in ('active_step', 'is_active') and self.active_step not in ('ext', ''))  # noqa
                or not self.is_active):
            return

        try:
            trace = self._get_ext_trace()
        except Exception:
            # NOTE: ignore error, but will be raised when clicking ANY of the export buttons
            for mark in ['trace', 'ext_lower', 'ext_upper']:
                self.marks[mark].clear()
        else:
            xs = range(len(trace.trace))
            self.marks['trace'].update_xy(xs,
                                          trace.trace,
                                          viewers=self.marks_viewers2d)
            self.marks['trace'].line_style = 'dashed'
            if self.ext_type_selected == 'Boxcar':
                self.marks['ext_lower'].update_xy(xs,
                                                  trace.trace-self.ext_width/2,
                                                  viewers=self.marks_viewers2d)
                self.marks['ext_upper'].update_xy(xs,
                                                  trace.trace+self.ext_width/2,
                                                  viewers=self.marks_viewers2d)
            else:
                for mark in ['ext_lower', 'ext_upper']:
                    self.marks[mark].clear()

        self._update_interactive_extract(event)

        self.active_step = 'ext'

        # TODO: remove this, the traitlet, and the row in spectral_extraction.vue
        # when specutils handles the warning/exception
        if self.ext_type_selected == 'Horne':
            inp_sp2d = self._get_ext_input_spectrum()
            self.ext_uncert_warn = isinstance(inp_sp2d.uncertainty, UnknownUncertainty)
        else:
            self.ext_uncert_warn = False

    def _set_create_kwargs(self, **kwargs):
        invalid_kwargs = [k for k in kwargs.keys() if not hasattr(self, k)]
        if len(invalid_kwargs):
            raise ValueError(f"{invalid_kwargs} are not valid attributes to pass as kwargs")

        for k, v in kwargs.items():
            setattr(self, k, v)

    def import_trace(self, trace):
        """
        Import the input parameters from an existing specreduce Trace object into the plugin.

        Parameters
        ----------
        trace : specreduce.tracing.Trace
            Trace object to import
        """
        if not isinstance(trace, tracing.Trace):  # pragma: no cover
            raise TypeError("trace must be a specreduce.tracing.Trace object")

        if isinstance(trace, tracing.FlatTrace):
            self.trace_type_selected = 'Flat'
            self.trace_pixel = trace.trace_pos
        elif isinstance(trace, tracing.FitTrace):
            self.trace_type_selected = trace.trace_model.__class__.__name__.strip('1D')
            self.trace_pixel = trace.guess
            self.trace_window = trace.window
            self.trace_bins = trace.bins
            self.trace_do_binning = True
            if hasattr(trace.trace_model, 'degree'):
                self.trace_order = trace.trace_model.degree
        elif isinstance(trace, tracing.ArrayTrace):  # pragma: no cover
            raise NotImplementedError(f"cannot import ArrayTrace into plugin.  Use viz.load instead")  # noqa
        else:  # pragma: no cover
            raise NotImplementedError(f"trace of type {trace.__class__.__name__} not supported")

    @with_spinner('trace_spinner')
    def export_trace(self, add_data=False, **kwargs):
        """
        Create a specreduce Trace object from the input parameters
        defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting trace to the application, according to the options
            defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'trace':
            self.update_marks(step='trace')

        if self.trace_trace_selected != 'New Trace':
            # then we're offsetting an existing trace
            # for FlatTrace, we can keep and expose a new FlatTrace (which has the advantage of
            # being able to load back into the plugin)
            orig_trace = self.trace_trace.selected_obj
            if isinstance(orig_trace, tracing.FlatTrace):
                trace = tracing.FlatTrace(self.trace_dataset.selected_obj,
                                          orig_trace.trace_pos+self.trace_offset)
            else:
                trace = tracing.ArrayTrace(self.trace_dataset.selected_obj,
                                           self.trace_trace.selected_obj.trace+self.trace_offset)

        elif self.trace_type_selected == 'Flat':
            trace = tracing.FlatTrace(self.trace_dataset.selected_obj,
                                      self.trace_pixel)

        elif self.trace_type_selected in _model_cls:
            trace_model = _model_cls[self.trace_type_selected](degree=self.trace_order)
            trace = tracing.FitTrace(self.trace_dataset.selected_obj,
                                     guess=self.trace_pixel,
                                     bins=int(self.trace_bins) if self.trace_do_binning else None,
                                     window=self.trace_window,
                                     peak_method=self.trace_peak_method_selected.lower(),
                                     trace_model=trace_model)

        else:
            raise NotImplementedError(f"trace_type={self.trace_type_selected} not implemented")

        if add_data:
            self.trace_add_results.add_results_from_plugin(trace, replace=False)

        return trace

    def vue_create_trace(self, *args):
        self.export_trace(add_data=True)

    def _get_bg_trace(self):
        if self.bg_type_selected == 'Manual':
            trace = tracing.FlatTrace(self.trace_dataset.get_selected_spectrum(
                                      use_display_units=True),
                                      self.bg_trace_pixel)
        elif self.bg_trace_selected == 'From Plugin':
            trace = self.export_trace(add_data=False)
        else:
            trace = self.bg_trace.get_selected_spectrum(use_disaply_units=True)

        return trace

    def import_bg(self, bg):
        """
        Import the input parameters from an existing specreduce Background object into the plugin.

        Parameters
        ----------
        bg : specreduce.background.Background
            Background object to import
        """
        if not isinstance(bg, background.Background):  # pragma: no cover
            raise TypeError("bg must be a specreduce.background.Background object")

        # TODO: should we detect/set the referenced dataset?
        trace = self._get_bg_trace()
        if len(bg.traces) == 2:
            # try to detect constant separation
            seps1 = bg.traces[0].trace - trace.trace
            seps2 = trace.trace - bg.traces[1].trace
            if np.all(seps1 == seps1[0]) and np.all(seps2 == seps1[0]):
                self.bg_type_selected = 'TwoSided'
                self.bg_separation = abs(int(seps1[0]))
            else:  # pragma: no cover
                raise NotImplementedError("backgrounds with custom traces not supported (could not detect common separation)")  # noqa
        elif len(bg.traces) == 1:
            # either one_sided or trace, let's see if its constant offset from the trace
            seps = bg.traces[0].trace - trace.trace
            if np.all(seps == seps[0]):
                self.bg_type_selected = 'OneSided'
                self.bg_separation = int(seps[0])
            else:  # pragma: no cover
                raise NotImplementedError("backgrounds with custom traces not supported (could not detect common separation)")  # noqa
        else:  # pragma: no cover
            raise NotImplementedError("backgrounds with more than 2 traces not supported")

        self.bg_width = bg.width

    @with_spinner('bg_spinner')
    def export_bg(self, **kwargs):
        """
        Create a specreduce Background object from the input parameters defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'bg':
            self.update_marks(step='bg')

        trace = self._get_bg_trace()

        if self.bg_type_selected == 'Manual':
            bg = background.Background(self.bg_dataset.get_selected_spectrum(
                                       use_display_units=True),
                                       [trace], width=self.bg_width,
                                       statistic=self.bg_statistic.selected.lower())
        elif self.bg_type_selected == 'OneSided':
            bg = background.Background.one_sided(self.bg_dataset.get_selected_spectrum(
                                                 use_display_units=True),
                                                 trace,
                                                 self.bg_separation,
                                                 width=self.bg_width,
                                                 statistic=self.bg_statistic.selected.lower())
        elif self.bg_type_selected == 'TwoSided':
            bg = background.Background.two_sided(self.bg_dataset.get_selected_spectrum(
                                                 use_display_units=True),
                                                 trace,
                                                 self.bg_separation,
                                                 width=self.bg_width,
                                                 statistic=self.bg_statistic.selected.lower())
        else:  # pragma: no cover
            raise NotImplementedError(f"bg_type={self.bg_type_selected} not implemented")

        return bg

    @with_spinner('bg_img_spinner')
    def export_bg_img(self, add_data=False, **kwargs):
        """
        Create a background 2D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting image to the application, according to the options
            defined in the plugin.
        """
        bg_spec = self.export_bg(**kwargs).bkg_image()

        if add_data:
            self.bg_add_results.add_results_from_plugin(bg_spec, replace=True)

        return bg_spec

    def vue_create_bg_img(self, *args):
        try:
            self.export_bg_img(add_data=True)
        except Exception as e:
            self.app.hub.broadcast(
                SnackbarMessage(f"Specreduce background failed with the following error: {repr(e)}",
                                color='error', sender=self)
            )

    @with_spinner('bg_spec_spinner')
    def export_bg_spectrum(self, add_data=False, **kwargs):
        """
        Create a background 1D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum to the application, according to the options
            defined in the plugin.
        """
        spec = self.export_bg(**kwargs).bkg_spectrum()

        if add_data:
            self.bg_spec_add_results.add_results_from_plugin(spec, replace=False)

        return spec

    def vue_create_bg_spec(self, *args):
        self.export_bg_spectrum(add_data=True)

    @with_spinner('bg_sub_spinner')
    def export_bg_sub(self, add_data=False, **kwargs):
        """
        Create a background-subtracted 2D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting image to the application, according to the options
            defined in the plugin.
        """
        bg_sub_spec = self.export_bg(**kwargs).sub_image()

        if add_data:
            self.bg_sub_add_results.add_results_from_plugin(bg_sub_spec, replace=True)

        return bg_sub_spec

    def vue_create_bg_sub(self, *args):
        self.export_bg_sub(add_data=True)

    def _get_ext_trace(self):
        if self.ext_trace_selected == 'From Plugin':
            return self.export_trace(add_data=False)
        else:
            return self.ext_trace.get_selected_spectrum(use_display_units=True)

    def _get_ext_input_spectrum(self):
        if self.ext_dataset_selected == 'From Plugin':
            return self.export_bg_sub(add_data=False)
        else:
            return self.ext_dataset.get_selected_spectrum(use_display_units=True)

    def import_extract(self, ext):
        """
        Import the input parameters from an existing specreduce extract object into the plugin.

        Parameters
        ----------
        ext : specreduce.extract.BoxcarExtract
            Extract object to import
        """
        if isinstance(ext, extract.BoxcarExtract):
            self.ext_type_selected = 'Boxcar'
            self.ext_width = ext.width
        elif isinstance(ext, extract.HorneExtract):
            self.ext_type_selected = 'Horne'
        else:  # pragma: no cover
            raise TypeError("ext must be a specreduce.extract.BoxcarExtract or specreduce.extract.HorneExtract object")  # noqa

    def export_extract(self, **kwargs):
        """
        Create a specreduce extraction object from the input parameters defined in the plugin.
        """
        self._set_create_kwargs(**kwargs)
        if len(kwargs) and self.active_step != 'ext':
            self.update_marks(step='ext')

        trace = self._get_ext_trace()
        inp_sp2d = self._get_ext_input_spectrum()

        if self.ext_type_selected == 'Boxcar':
            ext = extract.BoxcarExtract(inp_sp2d, trace, width=self.ext_width)
        elif self.ext_type_selected == 'Horne':
            spatial_profile = None
            if inp_sp2d.uncertainty is None:
                inp_sp2d.uncertainty = VarianceUncertainty(np.ones_like(inp_sp2d.data))

            if not hasattr(inp_sp2d.uncertainty, 'uncertainty_type'):
                inp_sp2d.uncertainty = StdDevUncertainty(inp_sp2d.uncert)

            if self.horne_ext_profile_selected == 'Self (interpolated)':

                # check inputs
                if self.self_prof_n_bins <= 0:
                    raise ValueError('`self_prof_n_bins` must be greater than 0.')
                if self.self_prof_interp_degree_x <= 0:
                    raise ValueError('`self_prof_interp_degree_x` must be greater than 0.')
                if self.self_prof_interp_degree_y <= 0:
                    raise ValueError('`self_prof_interp_degree_y` must be greater than 0.')

                # setup dict of interpolation options
                n_bins_interpolated_profile = self.self_prof_n_bins
                interp_degree = (self.self_prof_interp_degree_x, self.self_prof_interp_degree_y)
                spatial_profile = {'name': 'interpolated_profile',
                                   'n_bins_interpolated_profile': n_bins_interpolated_profile,
                                   'interp_degree': interp_degree}

            elif self.horne_ext_profile_selected == 'Gaussian':
                spatial_profile = 'gaussian'

            else:
                raise ValueError("Horne extraction profile must either be 'Gaussian' or 'Self (interpolated)'")  # noqa

            ext = extract.HorneExtract(inp_sp2d, trace, spatial_profile=spatial_profile)
        else:
            raise NotImplementedError(f"extraction type '{self.ext_type_selected}' not supported")  # noqa

        return ext

    @with_spinner('spinner')
    def export_extract_spectrum(self, add_data=False, **kwargs):
        """
        Create an extracted 1D spectrum from the input parameters defined in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum to the application, according to the options
            defined in the plugin.
        """
        extract = self.export_extract(**kwargs)
        spectrum = extract.spectrum

        if add_data:
            # TODO: eventually generalize this logic into add_results_from_plugin
            if not len(self.marks_viewers1d):
                # no spectrum1d viewer, create one now and set the default viewer
                viewer_ref = self.app.return_unique_name('1D Spectrum',
                                                         typ='viewer')
                self.app._on_new_viewer(NewViewerMessage(Spectrum1DViewer,
                                                         data=None,
                                                         sender=self.app),
                                        vid=viewer_ref, name=viewer_ref,
                                        open_data_menu_if_empty=False)
                self.ext_add_results.viewer = viewer_ref

            self.ext_add_results.add_results_from_plugin(spectrum, replace=False)

        return spectrum

    def vue_extract_spectrum(self, *args):
        self.export_extract_spectrum(add_data=True)



from traitlets import Bool, List, Unicode, observe
from astropy.io import fits
from astropy import units as u
from astropy.wcs import WCS
from specutils import Spectrum

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.template_mixin import AutoTextField, SelectFileExtensionComponent
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.utils import standardize_metadata, PRIHDR_KEY


__all__ = ['Spectrum2DImporter']


def hdu_is_valid(item):
    """
    Check if the HDU is valid to be imported as a 2D Spectrum.

    Parameters
    ----------
    hdu : `astropy.io.fits.hdu.base.HDUBase`
        The HDU to check.

    Returns
    -------
    bool
        True if the HDU is a valid light curve HDU, False otherwise.
    """
    hdu = item.get('obj')
    return (len(getattr(hdu, 'shape', [])) == 2
            and ('DISPAXIS' in hdu.header
                 or hdu.header.get('CTYPE1', '') == 'WAVE'
                 or hdu.header.get('EXTNAME', '') == 'FLUX'))


@loader_importer_registry('2D Spectrum')
class Spectrum2DImporter(BaseImporterToDataCollection):
    template_file = __file__, "./spectrum2d.vue"
    parser_preference = ['fits', 'specutils.Spectrum']

    auto_extract = Bool(True).tag(sync=True)

    ext_data_label_value = Unicode().tag(sync=True)
    ext_data_label_default = Unicode().tag(sync=True)
    ext_data_label_auto = Bool(True).tag(sync=True)
    ext_data_label_invalid_msg = Unicode().tag(sync=True)

    # HDUList-specific options
    input_hdulist = Bool(False).tag(sync=True)
    extension_items = List().tag(sync=True)
    extension_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz2d':
            self.data_label_default = '2D Spectrum'

        self.ext_data_label = AutoTextField(self,
                                            'ext_data_label_value',
                                            'ext_data_label_default',
                                            'ext_data_label_auto',
                                            'ext_data_label_invalid_msg')

        self.input_hdulist = isinstance(self.input, fits.HDUList)
        if self.input_hdulist:
            ext_options = [{'label': f"{index}: {hdu.name}",
                            'name': hdu.name,
                            'ver': hdu.ver,
                            'name_ver': f"{hdu.name},{hdu.ver}",
                            'index': index,
                            'obj': hdu}
                           for index, hdu in enumerate(self.input)]
            self.extension = SelectFileExtensionComponent(self,
                                                          items='extension_items',
                                                          selected='extension_selected',
                                                          manual_options=ext_options,
                                                          filters=[hdu_is_valid])

    @property
    def user_api(self):
        expose = ['auto_extract', 'ext_data_label']
        if self.input_hdulist:
            expose += ['extension']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz2d'):
            # NOTE: temporary during deconfig process
            return False
        if not ((isinstance(self.input, Spectrum)
                 and self.input.flux.ndim == 2) or
                (isinstance(self.input, fits.HDUList)
                 and len([hdu for hdu in self.input if hdu_is_valid({'obj': hdu})]))):  # noqa
            return False
        try:
            self.output
        except Exception:
            return False
        return True

    @property
    def default_viewer_reference(self):
        # returns the registry name of the default viewer
        # only used if `show_in_viewer=True` and no existing viewers can accept the data
        return 'spectrum-2d-viewer'

    @observe('data_label_value')
    def _data_label_changed(self, msg={}):
        self.ext_data_label_default = f"{self.data_label_value} (auto-ext)"

    @property
    def output(self):
        if not self.input_hdulist:
            return self.input

        hdulist = self.input
        hdu = self.extension.selected_obj
        data = hdu.data
        header = hdu.header
        metadata = standardize_metadata(header)
        if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
            metadata[PRIHDR_KEY] = standardize_metadata(hdulist[0].header)
        wcs = WCS(header, hdulist)
        if data.shape[0] > data.shape[1]:
            data = data.T
            self.app.hub.broadcast(SnackbarMessage(
                f"Transposed input data to {data.shape}",
                sender=self, color="warning"))
        if wcs.array_shape[0] > wcs.array_shape[1]:
            wcs = wcs.swapaxes(0, 1)

        try:
            data_unit = u.Unit(header['BUNIT'])
        except Exception:
            data_unit = u.count

        try:
            if wcs.world_axis_physical_types == [None, None]:
                # This may be a JWST file with WCS stored in ASDF
                if 'ASDF' in hdulist:
                    try:
                        from stdatamodels import asdf_in_fits
                        tree = asdf_in_fits.open(hdulist).tree
                        if 'meta' in tree and 'wcs' in tree['meta']:
                            wcs = tree["meta"]["wcs"][0]
                        else:
                            wcs = None
                    except ValueError:
                        wcs = None
                else:
                    wcs = None
            return Spectrum(flux=data * data_unit, meta=metadata, wcs=wcs, spectral_axis_index=1)
        except ValueError:
            # In some cases, the above call to Spectrum will fail if no
            # spectral axis is found in the WCS. Even without a spectral axis,
            # the Spectrum.read parser may work, so we try that next.
            # If that also fails, then drop the WCS.
            try:
                Spectrum.read(self._resolver())
            except Exception:
                # specutils.Spectrum > Spectrum2D would fail, so use no WCS
                return Spectrum(flux=data * data_unit, meta=metadata)
            else:
                # raising an error here will allow using specutils.Spectrum > Spectrum2D
                raise

    def __call__(self, show_in_viewer=True):
        # get a copy of both of these before additional data entries changes defaults
        data_label = self.data_label_value
        ext_data_label = self.ext_data_label_value

        super().__call__(show_in_viewer=show_in_viewer)

        if not self.auto_extract:
            return

        try:
            spext = self.app.get_tray_item_from_name('spectral-extraction-2d')
            ext = spext._extract_in_new_instance(dataset=data_label,
                                                 add_data=False)
        except Exception:
            ext = None
            msg = SnackbarMessage(
                "Automatic spectrum extraction failed. See the spectral extraction"
                " plugin to perform a custom extraction",
                color='error', sender=self, timeout=10000)
        else:
            msg = SnackbarMessage(
                "The extracted 1D spectrum was generated automatically."
                " See the spectral extraction plugin for details or to"
                " perform a custom extraction.",
                color='warning', sender=self, timeout=10000)
        self.app.hub.broadcast(msg)

        if ext is not None:
            self.add_to_data_collection(ext, ext_data_label, show_in_viewer=False)
            self.load_into_viewer(ext_data_label, "spectrum-1d-viewer")



from functools import cached_property

from glue.core import BaseData
from glue_jupyter.bqplot.image import BqplotImageView
import numpy as np

from jdaviz.core.registries import viewer_registry
from jdaviz.configs.cubeviz.plugins.mixins import WithSliceIndicator, WithSliceSelection
from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin
from jdaviz.configs.specviz.plugins.viewers import Spectrum1DViewer
from jdaviz.core.freezable_state import FreezableBqplotImageViewerState
from jdaviz.configs.cubeviz.plugins.cube_listener import MINVOL
from jdaviz.core.sonified_layers import (SonifiedDataLayerArtist,
                                         SonifiedLayerStateWidget,
                                         SonifiedLayerState)


__all__ = ['CubevizImageView', 'CubevizProfileView']


@viewer_registry("cubeviz-image-viewer", label="Image 2D (Cubeviz)")
class CubevizImageView(JdavizViewerMixin, WithSliceSelection, BqplotImageView):
    # categories: zoom resets, (zoom, pan), subset, select tools, shortcuts
    # NOTE: zoom and pan are merged here for space consideration and to avoid
    # overflow to second row when opening the tray
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:pixelboxzoommatch', 'jdaviz:boxzoom',
                     'jdaviz:pixelpanzoommatch', 'jdaviz:panzoom'],
                    ['bqplot:truecircle', 'bqplot:rectangle', 'bqplot:ellipse',
                     'bqplot:circannulus'],
                    ['jdaviz:spectrumperspaxel'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = None
    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # provide reference from state back to viewer to use for zoom syncing
        self.state._set_viewer(self)

        self._subscribe_to_layers_update()
        self.state.add_callback('reference_data', self._initial_x_axis)

        self.add_event_callback(self._viewer_mouse_event, events=['mousemove',
                                                                  'mouseleave',
                                                                  'mouseenter'])

        # Hide axes by default
        self.state.show_axes = False

        self.volume_level = None

        self.data_menu._obj.dataset.add_filter('is_cube_or_image')

        self.layer_volume = {}
        self.same_pix = None
        self._cached_properties = ['combined_sonified_grid']
        self.sonified_layers_enabled = []

        self._layer_style_widget_cls[SonifiedDataLayerArtist] = SonifiedLayerStateWidget

    @property
    def _sonify_plugin(self):
        if self.jdaviz_helper is not None and 'Sonify Data' in self.jdaviz_helper.plugins:
            return self.jdaviz_helper.plugins['Sonify Data']._obj
        else:
            return None

    @property
    def _default_spectrum_viewer_reference_name(self):
        return self.jdaviz_helper._default_spectrum_viewer_reference_name

    @property
    def _default_flux_viewer_reference_name(self):
        return self.jdaviz_helper._default_flux_viewer_reference_name

    @property
    def _default_uncert_viewer_reference_name(self):
        return self.jdaviz_helper._default_uncert_viewer_reference_name

    def _initial_x_axis(self, *args):
        # Make sure that the x_att is correct on data load
        ref_data = self.state.reference_data
        if ref_data and ref_data.ndim == 3:
            for att_name in ["Right Ascension", "RA", "Galactic Longitude"]:
                if att_name in ref_data.component_ids():
                    x_att = att_name
                    self.state.x_att_world = ref_data.id[x_att]
                    break

    def set_plot_axes(self):
        self.figure.axes[1].tick_format = None
        self.figure.axes[0].tick_format = None

        self.figure.axes[1].label = "y: pixels"
        self.figure.axes[0].label = "x: pixels"

        # Make it so y axis label is not covering tick numbers.
        self.figure.axes[1].label_offset = "-50"

    def data(self, cls=None):
        return [layer_state.layer  # .get_object(cls=cls or self.default_class)
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                isinstance(layer_state.layer, BaseData)]

    @cached_property
    def combined_sonified_grid(self):

        compiled_coords = {}
        for k, v in self._sonify_plugin.data_lookup.items():
            if k not in self.sonified_layers_enabled:
                continue
            # Each (x, y) coordinate corresponds to a different sound for each layer.
            # These sounds can be combined together and played by setting cbuff to True.
            # TODO: is there a better way to combine sounds or normalize them?
            # TODO: apply 1/N or 1/N**0.5 normalisation per layer for N layers?
            for coord, sound_array in v.items():
                if coord in compiled_coords:
                    compiled_coords[coord] += ((sound_array * (int(self.layer_volume[k]) / 100)).
                                               astype(int))
                else:
                    compiled_coords[coord] = ((sound_array * (int(self.layer_volume[k]) / 100)).
                                              astype(int))
        return compiled_coords

    def recalculate_combined_sonified_grid(self, event=None):
        self.layer_volume = {}
        self.sonified_layers_enabled = []

        for layer in self.state.layers:
            if not isinstance(layer, SonifiedLayerState):
                continue

            self.layer_volume[layer.layer.label] = layer.volume
            self.sonified_layers_enabled += ([layer.layer.label] if
                                             getattr(layer, 'audible', False) else [])  # noqa

        # Need to force an update of the layer icons since
        # audible is a state attribute, not a layer artist attribute
        self.jdaviz_app.state.layer_icons.notify_all()

        if 'combined_sonified_grid' in self.__dict__:
            del self.__dict__['combined_sonified_grid']
        self.combined_sonified_grid

    def update_volume_level(self, level):
        if self._sonify_plugin is None or not self._sonify_plugin.sonified_cube:
            return
        self.volume_level = level
        self._sonify_plugin.sonified_cube.atten_level = int(1/np.clip((level/100.)**2, MINVOL, 1))

    def _viewer_mouse_event(self, data):
        if data['event'] in ('mouseleave', 'mouseenter') or not self.sonified_layers_enabled:
            if self._sonify_plugin is not None:
                self._sonify_plugin.stop_stream()
            return
        if len(self.jdaviz_app.data_collection) < 1:
            return

        # Extract data coordinates - these are pixels in the reference image
        x = np.floor(data['domain']['x'])
        y = np.floor(data['domain']['y'])

        if x is None or y is None or x < 0 or y < 0:  # Out of bounds
            return

        if self.same_pix is None:
            self.same_pix = (x, y)
        elif (x, y) == self.same_pix:
            return

        if (not self._sonify_plugin.sonified_cube or
                not hasattr(self._sonify_plugin.sonified_cube, 'newsig') or
                not hasattr(self._sonify_plugin.sonified_cube, 'sigcube')):
            return
        self._sonify_plugin.start_stream()
        self._sonify_plugin.update_sonified_cube_with_coord(self, (x, y))

    def get_data_layer_artist(self, layer=None, layer_state=None):
        if 'Sonified' in layer.meta:
            cls = SonifiedDataLayerArtist
            return self.get_layer_artist(cls, layer=layer)
        else:
            return super().get_data_layer_artist(layer, layer_state)


@viewer_registry("cubeviz-profile-viewer", label="Profile 1D (Cubeviz)")
class CubevizProfileView(Spectrum1DViewer, WithSliceIndicator):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom', 'jdaviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['jdaviz:panzoom', 'jdaviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['bqplot:xrange'],
                    ['jdaviz:selectslice', 'jdaviz:selectline'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default_tool_priority', ['jdaviz:selectslice'])
        super().__init__(*args, **kwargs)

    @property
    def _default_flux_viewer_reference_name(self):
        return self.jdaviz_helper._default_flux_viewer_reference_name

    def add_subset(self, subset, *args, **kwargs):
        # The cubeviz profile viewer does not show the spectra/profiles based
        # on the subsets of the cubes, but based on extracted datasets derived
        # from those subsets. We can ignore all subsets that have a subset
        # state that was defined from the image, and we recognize these by
        # looking for subset states that are defined based on two attributes.
        if len(subset.subset_state.attributes) == 2:
            return False
        else:
            return super().add_subset(subset, *args, **kwargs)



from functools import cached_property
from skimage.io import imread
from skimage.color import rgb2gray, rgba2rgb
from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


__all__ = ['JPGPNGParser']


@loader_parser_registry('jpgpng')
class JPGPNGParser(BaseParser):

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz2d', 'lcviz', 'imviz'):
            # NOTE: temporary during deconfig process
            return False
        return isinstance(self.input, str) and self.input.endswith(('.jpg', '.jpeg', '.png'))

    @cached_property
    def output(self):
        im = imread(self.input)
        if im.shape[2] == 4:
            pf = rgb2gray(rgba2rgb(im))
        else:  # Assume RGB
            pf = rgb2gray(im)
        return pf[::-1, :]  # Flip it



from traitlets import Bool, Unicode, observe
from glue.core.message import (DataCollectionAddMessage,
                               DataCollectionDeleteMessage)

from jdaviz.core.events import NewViewerMessage, SnackbarMessage
from jdaviz.core.registries import viewer_registry
from jdaviz.core.template_mixin import PluginTemplateMixin, AutoTextField
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.utils import standardize_metadata

__all__ = ['BaseImporter', 'BaseImporterToDataCollection', 'BaseImporterToPlugin']


vid_map = {'spectrum-1d-viewer': '1D Spectrum',
           'spectrum-2d-viewer': '2D Spectrum',
           'imviz-image-viewer': 'Image'}


class BaseImporter(PluginTemplateMixin):
    # preference order of parsers, by registry name.  If empty, the first found match will
    # be used by default.  If not empty, the first match in the list will be used (including
    # over any parsers not included in the list).  If not empty but no valid parsers are in
    # the list, the first remaining match will be used.
    parser_preference = []

    def __init__(self, app, resolver, input, **kwargs):
        self._input = input
        self._resolver = resolver
        super().__init__(app, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    @property
    def is_valid(self):
        # override by subclass
        return False  # pragma: nocover

    @property
    def resolver(self):
        return self._resolver

    @property
    def input(self):
        return self._input

    @property
    def output(self):
        # override by subclass
        return self.input

    @property
    def target(self):
        raise NotImplementedError("Importer subclass must implement target")  # pragma: nocover

    def __call__(self):
        # override by subclass - should act on self.output and load into jdaviz
        raise NotImplementedError("Importer subclass must implement __call__")  # pragma: nocover

    @property
    def user_api(self):
        return ImporterUserApi(self)


class BaseImporterToDataCollection(BaseImporter):
    data_label_value = Unicode().tag(sync=True)
    data_label_default = Unicode().tag(sync=True)
    data_label_auto = Bool(True).tag(sync=True)
    data_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, app, resolver, input, **kwargs):
        super().__init__(app, resolver, input, **kwargs)
        self.data_label_default = self._registry_label
        self.data_label = AutoTextField(self, 'data_label_value',
                                        'data_label_default',
                                        'data_label_auto',
                                        'data_label_invalid_msg')

        self.hub.subscribe(self, DataCollectionAddMessage,
                           handler=lambda _: self._on_label_changed())
        self.hub.subscribe(self, DataCollectionDeleteMessage,
                           handler=lambda _: self._on_label_changed())
        self.observe(self._on_label_changed, 'data_label_value')
        self._on_label_changed()

    @property
    def default_viewer_reference(self):
        raise NotImplementedError("Importer subclass must implement default_viewer_reference")  # noqa pragma: nocover

    @property
    def ignore_viewers_with_cls(self):
        return ()

    @property
    def default_data_label_from_resolver(self):
        return self._resolver.default_label

    @property
    def default_viewer_label(self):
        return vid_map.get(self.default_viewer_reference, self.default_viewer_reference)

    @property
    def target(self):
        return {'type': 'viewer',
                'icon': 'mdi-window-maximize',
                'label': self.default_viewer_label}

    def _on_label_changed(self, msg={}):
        if not len(self.data_label_value.strip()):
            # strip will raise the same error for a label of all spaces
            self.data_label_invalid_msg = 'data_label must be provided'
            return

        # ensure the default label is unique for the data-collection
        self.data_label_default = self.app.return_unique_name(self.data_label_default)

        for data in self.app.data_collection:
            if self.data_label_value == data.label:
                self.data_label_invalid_msg = 'data_label already in use'
                return

        self.data_label_invalid_msg = ''

    @observe('data_label_invalid_msg')
    def _set_import_disabled(self, change={}):
        self.resolver.import_disabled = len(self.data_label_invalid_msg) > 0

    def load_into_viewer(self, data_label, default_viewer_reference=None):
        added = 0
        for viewer in self.app._jdaviz_helper.viewers.values():
            if isinstance(viewer._obj, self.ignore_viewers_with_cls):
                continue
            if data_label in viewer.data_menu.data_labels_unloaded:
                added += 1
                viewer.data_menu.add_data(data_label)
            elif data_label in viewer.data_menu.data_labels_loaded:
                # was already loaded, increment count to avoid creating a new viewer
                added += 1
        if added == 0:
            if self.app.config not in ('deconfigged', 'lcviz'):
                # do not add additional viewers
                msg = SnackbarMessage(
                    "Data units are incompatible with viewer units, unload all data from viewer to add",  # noqa
                    color='error', sender=self, timeout=10000)
                self.app.hub.broadcast(msg)
                return
            if default_viewer_reference is None:
                default_viewer_reference = self.default_viewer_reference
                default_viewer_label = self.default_viewer_label
            else:
                default_viewer_label = vid_map.get(default_viewer_reference,
                                                   default_viewer_reference)
            default_viewer_label = self.app.return_unique_name(default_viewer_label,
                                                               typ='viewer')

            viewer_dict = viewer_registry.members.get(default_viewer_reference)
            viewer_cls = viewer_dict.get('cls')
            self.app._on_new_viewer(NewViewerMessage(viewer_cls, data=None, sender=self.app),
                                    vid=default_viewer_label,
                                    name=default_viewer_label,
                                    open_data_menu_if_empty=False)
            viewer = self.app._jdaviz_helper.viewers.get(default_viewer_label)
            viewer.data_menu.add_data(data_label)

    def add_to_data_collection(self, data, data_label=None,
                               parent=None, show_in_viewer=True, cls=None):
        if data_label is None:
            data_label = self.data_label_value.strip()
        if hasattr(data, 'meta'):
            try:
                data.meta = standardize_metadata(data.meta)
            except TypeError:
                pass
        self.app.add_data(data, data_label=data_label)
        if parent is not None:
            self.app._set_assoc_data_as_child(data_label, parent)
        # store the original input class so that get_data can default to the
        # same class as the input
        cls = cls if cls is not None else data.__class__
        self.app.data_collection[data_label]._native_data_cls = cls
        self.app.data_collection[data_label]._importer = self.__class__.__name__
        if show_in_viewer:
            self.load_into_viewer(data_label)

    def __call__(self, show_in_viewer=True):
        if self.data_label_invalid_msg:
            raise ValueError(self.data_label_invalid_msg)
        self.add_to_data_collection(self.output, show_in_viewer=show_in_viewer)


class BaseImporterToPlugin(BaseImporter):
    @property
    def default_plugin(self):
        raise NotImplementedError("Importer subclass must implement default_plugin")  # noqa pragma: nocover

    @property
    def target(self):
        return {'type': 'plugin',
                'icon': 'mdi-toy-brick-outline',
                'label': self.default_plugin}

    @property
    def has_default_plugin(self):
        return self.default_plugin in self.app._jdaviz_helper.plugins



from jdaviz.configs.specviz.plugins.viewers import Spectrum1DViewer
from jdaviz.core.viewer_creators import BaseViewerCreator
from jdaviz.core.registries import viewer_creator_registry


__all__ = ['Spectrum1DViewerCreator']


@viewer_creator_registry('1D Spectrum', overwrite=True)
class Spectrum1DViewerCreator(BaseViewerCreator):
    template_file = __file__, "../base_viewer_creator.vue"

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        self.dataset.filters = ['is_spectrum']
        self.viewer_label_default = '1D Spectrum'  # TODO: append suffix if taken

    @property
    def viewer_class(self):
        return Spectrum1DViewer



from traitlets import Bool, Unicode, observe
from urllib.parse import urlparse
import os
from functools import cached_property

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import BaseResolver
from jdaviz.core.user_api import LoaderUserApi
from jdaviz.utils import download_uri_to_path, get_cloud_fits


__all__ = ['URLResolver']


@loader_resolver_registry('url')
class URLResolver(BaseResolver):
    template_file = __file__, "url.vue"
    default_input = 'url'

    url = Unicode("").tag(sync=True)
    url_scheme = Unicode("").tag(sync=True)
    cache = Bool(True).tag(sync=True)
    local_path = Unicode("").tag(sync=True)
    timeout = FloatHandleEmpty(10).tag(sync=True)

    def __init__(self, *args, **kwargs):
        self.local_path = os.curdir
        super().__init__(*args, **kwargs)

    @property
    def user_api(self):
        return LoaderUserApi(self, expose=['url', 'cache', 'local_path', 'timeout'])

    @property
    def is_valid(self):
        return self.url_scheme in ['http', 'https', 'mast', 'ftp', 's3']

    @property
    def default_label(self):
        # Use the last part of the URL as the default label, if available.
        # Otherwise, return None.
        if self.url.strip():
            return os.path.splitext(os.path.basename(self.url.strip()))[0]
        return None

    @observe('url', 'cache', 'timeout')
    def _on_url_changed(self, change):
        self.url_scheme = urlparse(self.url.strip()).scheme

        # Clear the cached property to force re-download
        # or otherwise read from local file cache.
        if '_uri_output_file' in self.__dict__ and change['name'] in ('url', 'cache'):
            del self._uri_output_file

        self._update_format_items()

    @cached_property
    def _uri_output_file(self):
        if self.url_scheme == 's3':
            return get_cloud_fits(self.url.strip())
        return download_uri_to_path(self.url.strip(), cache=self.cache,
                                    local_path=self.local_path, timeout=self.timeout)

    def __call__(self):
        return self._uri_output_file



import numpy as np
from specutils import Spectrum

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection


__all__ = ['SpectrumImporter']


@loader_importer_registry('1D Spectrum')
class SpectrumImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz':
            self.data_label_default = '1D Spectrum'
        elif self.app.config == 'specviz2d':
            self.data_label_default = '1D Spectrum'
        else:
            self.data_label_default = '1D Spectrum'

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz', 'specviz2d', 'cubeviz'):
            # cubeviz allowed for cubeviz.specviz.load_data support
            # NOTE: temporary during deconfig process
            return False
        return isinstance(self.input, Spectrum) and self.input.flux.ndim == 1

    @property
    def default_viewer_reference(self):
        # returns the registry name of the default viewer
        # only used if `show_in_viewer=True` and no existing viewers can accept the data
        return 'spectrum-1d-viewer'

    @property
    def output(self):
        # if the entire uncert. array is Nan and the data is not, model fitting won't
        # work (more generally, if uncert[i] is nan/inf and flux[i] is not, fitting will
        # fail, but just deal with the all nan case here since it is straightforward).
        # set uncerts. to None if they are all nan/inf, and display a warning message.
        data = self.input
        if data.uncertainty is not None:
            uncerts_finite = np.isfinite(data.uncertainty.array)
            if not np.any(uncerts_finite):
                data.uncertainty = None
                set_nans_to_none = True

                if set_nans_to_none:
                    # alert user that we have changed their all-nan uncertainty array to None
                    msg = 'All uncertainties are nonfinite, replacing with uncertainty=None.'
                    self.app.hub.broadcast(SnackbarMessage(msg, color="warning", sender=self.app))
        return data



from traitlets import List

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin
from jdaviz.core.user_api import PluginUserApi

try:
    from jdaviz import __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

__all__ = ['Logger']


@tray_registry('logger', label="Logger",
               category='core', sidebar='info', subtab=2)
class Logger(PluginTemplateMixin):
    """Show snackbar messages in a logger UI."""
    template_file = __file__, "logger.vue"

    history = List().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Access history of logger messages.'

    @property
    def user_api(self):
        expose = ['history']
        return PluginUserApi(self, expose=expose)



import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.wcs import WCS
from gwcs import WCS as GWCS
from specutils import SpectralRegion, Spectrum
from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.loaders.resolvers import find_matching_resolver
from jdaviz.utils import cached_uri


def test_loaders_registry(specviz_helper):
    # built-in loaders: file, url, etc
    assert len(specviz_helper.loaders) == len(loader_resolver_registry.members)


def test_open_close(specviz_helper):
    specviz_helper.app.state.dev_loaders = True

    assert specviz_helper.app.state.drawer_content == ''

    loader = specviz_helper.loaders['file']
    loader.open_in_tray()
    assert specviz_helper.app.state.drawer_content == 'loaders'
    loader.close_in_tray()
    assert specviz_helper.app.state.drawer_content == 'loaders'
    loader.close_in_tray(close_sidebar=True)
    assert specviz_helper.app.state.drawer_content == ''

    subset_plg = specviz_helper.plugins['Subset Tools']
    subset_plg.open_in_tray()
    assert specviz_helper.app.state.drawer_content == 'plugins'

    assert subset_plg._obj.loader_panel_ind is None  # loader panel not open
    subset_plg._obj.loaders['url'].open_in_tray()
    assert subset_plg._obj.loader_panel_ind == 0  # loader panel open
    subset_plg._obj.loaders['url'].close_in_tray()
    assert subset_plg._obj.loader_panel_ind is None  # loader panel not open
    subset_plg._obj.loaders['url'].close_in_tray(close_sidebar=True)
    assert specviz_helper.app.state.drawer_content == ''


def test_resolver_matching(specviz_helper):
    sp = Spectrum(spectral_axis=np.array([1, 2, 3])*u.nm,
                  flux=np.array([1, 2, 3])*u.Jy)

    res_sp = find_matching_resolver(specviz_helper.app, sp)
    assert res_sp._obj._registry_label == 'object'
    assert res_sp.format == '1D Spectrum'

    specviz_helper._load(sp)
    assert len(specviz_helper.app.data_collection) == 1


def test_trace_importer(specviz2d_helper, spectrum2d):
    specviz2d_helper._load(spectrum2d, format='2D Spectrum')

    trace = specviz2d_helper.plugins['2D Spectral Extraction'].export_trace()

    res_sp = find_matching_resolver(specviz2d_helper.app, trace)
    assert res_sp._obj._registry_label == 'object'
    assert res_sp.format == 'Trace'

    # import through loader API
    ldr = specviz2d_helper.loaders['object']
    ldr.object = trace
    assert ldr.format == 'Trace'
    ldr.importer.data_label = 'Trace 1'
    ldr.importer()
    assert specviz2d_helper.app.data_collection[-1].label == 'Trace 1'

    # import through load method
    specviz2d_helper._load(trace, data_label='Trace 2')
    assert specviz2d_helper.app.data_collection[-1].label == 'Trace 2'


def test_markers_specviz2d_unit_conversion(specviz2d_helper, spectrum2d):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.AA)
    specviz2d_helper.load_data(spectrum2d)


@pytest.mark.remote_data
@pytest.mark.filterwarnings(r"ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_fits_spectrum2d(deconfigged_helper):
    uri = cached_uri('mast:jwst/product/jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d.fits')
    if 'mast' in uri:
        ldr = deconfigged_helper.loaders['url']
        ldr.cache = True
        ldr.url = uri
    else:
        ldr = deconfigged_helper.loaders['file']
        ldr.filepath = uri

    # Default is Image but the test switches to 2D Spectrum
    # since this file type is not yet supported by the image loader
    assert ldr.format == 'Image'
    assert ldr.importer._obj.input_has_extensions is True

    ldr.format = '2D Spectrum'

    ldr.importer()

    # ensure get_data works, retrieves a Spectrum1D object, and has spectral WCS attached correctly
    sp2d = deconfigged_helper.get_data('jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d')  # noqa
    assert isinstance(sp2d, Spectrum)
    assert str(sp2d.spectral_axis.unit) == 'um'

    sp1d = deconfigged_helper.get_data('jw02123-o001_v000000353_nirspec_f170lp-g235h_s2d (auto-ext)')  # noqa
    assert isinstance(sp1d, Spectrum)
    assert str(sp1d.spectral_axis.unit) == 'um'


@pytest.mark.remote_data
def test_resolver_url(deconfigged_helper):
    loader = deconfigged_helper.loaders['url']

    # no url, no valid formats
    assert len(loader.format.choices) == 0

    # non-valid input
    loader.url = 'not-valid-url'
    assert len(loader.format.choices) == 0

    # s3 input
    loader.url = "s3://stpubdata/jwst/public/jw02727/L3/t/o002/jw02727-o002_t062_nircam_clear-f277w_i2d.fits"  # noqa: E501
    assert loader._obj.url_scheme == 's3'
    assert len(loader.format.choices) > 0

    # https valid input
    loader.url = 'https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits'  # noqa
    assert len(loader.format.choices) == 4  # may change with future importers
    assert loader.format.selected == 'Image'  # default may change with future importers

    # test target filtering
    assert len(loader.target.choices) > 1
    assert loader.target.selected == 'Any'
    loader.target = '1D Spectrum'
    assert len(loader.format.choices) == 2  # may change with future importers
    assert loader.format == '1D Spectrum List'  # default may change with future importers
    assert loader.importer.data_label == 'exnkul627fcuhy5akf2gswytud5tazmw'  # noqa

    loader.target = 'Any'
    assert len(loader.format.choices) == 4
    loader.format = '2D Spectrum'
    assert loader.importer.data_label == 'exnkul627fcuhy5akf2gswytud5tazmw'  # noqa

    assert len(deconfigged_helper.app.data_collection) == 0
    assert len(deconfigged_helper.viewers) == 0

    loader.importer()

    # 2D spectrum and auto-extracted 1D spectrum
    assert len(deconfigged_helper.app.data_collection) == 2
    assert len(deconfigged_helper.viewers) == 2

    with pytest.raises(ValueError, match="Failed query for URI"):
        # NOTE: this test will attempt to reach out to MAST via astroquery
        # even if cache is available.
        deconfigged_helper.load('mast:invalid')


def test_invoke_from_plugin(specviz_helper, spectrum1d, tmp_path):
    s = SpectralRegion(5*u.um, 6*u.um)
    local_path = str(tmp_path / 'spectral_region.ecsv')
    s.write(local_path)

    specviz_helper.load_data(spectrum1d)

    loader = specviz_helper.plugins['Subset Tools'].loaders['file']

    assert len(loader.format.choices) == 0
    loader.filepath = local_path
    assert len(loader.format.choices) > 0

    loader.importer()


def test_load_image_mult_sci_extension(imviz_helper):
    # test loading an image with multiple SCI extensions and
    # ensure that automatic parenting logic is handled correctly
    arr = np.zeros((2, 2), dtype=np.float32)
    hdul = fits.HDUList([fits.PrimaryHDU(),
                        fits.ImageHDU(arr, name='SCI', ver=1),
                        fits.ImageHDU(arr, name='ERR', ver=1),
                        fits.ImageHDU(arr, name='SCI', ver=2),
                        fits.ImageHDU(arr, name='ERR', ver=2)
                         ])

    # imviz_helper._load(hdul, extension=('SCI,1', 'SCI,2', 'ERR,2'))
    imviz_helper.load_data(hdul, ext=('SCI,1', 'SCI,2', 'ERR,2'))

    assert len(imviz_helper.app.data_collection) == 3
    assert [d.label for d in imviz_helper.app.data_collection] == ['Image[SCI,1]', 'Image[SCI,2]', 'Image[ERR,2]']  # noqa

    assert imviz_helper.app._get_assoc_data_children('Image[SCI,1]') == []
    assert imviz_helper.app._get_assoc_data_children('Image[SCI,2]') == ['Image[ERR,2]']
    assert imviz_helper.app._get_assoc_data_parent('Image[ERR,2]') == 'Image[SCI,2]'


def test_loaders_extension_select(imviz_helper):
    # tests internal logic of SelectFileExtensionComponent
    arr = np.zeros((2, 2), dtype=np.float32)
    hdul = fits.HDUList([fits.PrimaryHDU(),
                        fits.ImageHDU(arr, name='SCI', ver=1),
                        fits.ImageHDU(arr, name='ERR', ver=1),
                        fits.ImageHDU(arr, name='SCI', ver=2),
                        fits.ImageHDU(arr, name='ERR', ver=2)
                         ])

    ldr = imviz_helper.loaders['object']
    ldr.object = hdul

    # test setting extension by index: [name, ver], [name,ver], and name,ver
    ldr.importer.extension = ['1: [SCI,1]', '[SCI,2]', 'ERR,2']
    assert ldr.importer.extension.selected == ['1: [SCI,1]',
                                               '3: [SCI,2]',
                                               '4: [ERR,2]']
    assert ldr.importer.extension.selected_index == [1, 3, 4]

    # if not providing ver, first match will be used
    ldr.importer.extension = 'SCI'
    assert ldr.importer.extension.selected == ['1: [SCI,1]']

    # case of providing index
    ldr.importer.extension = [1, 3]
    assert ldr.importer.extension.selected == ['1: [SCI,1]', '3: [SCI,2]']


@pytest.mark.remote_data
@pytest.mark.parametrize(
    ('gwcs_to_fits_sip', 'expected_cls'),
    ((True, WCS), (False, GWCS),), ids=('True-WCS', 'False-GWCS'))
@pytest.mark.filterwarnings("ignore:Some non-standard WCS keywords were excluded")
def test_gwcs_to_fits_sip(gwcs_to_fits_sip, expected_cls, deconfigged_helper):
    """Test gwcs_to_fits_sip through the importer API."""
    ldr = deconfigged_helper.loaders['url']
    ldr.url = 'https://data.science.stsci.edu/redirect/JWST/jwst-data_analysis_tools/imviz_test_data/jw00042001001_01101_00001_nrcb5_cal.fits'  # noqa
    ldr.importer.gwcs_to_fits_sip = gwcs_to_fits_sip

    ldr.importer()

    data = deconfigged_helper.app.data_collection[0]
    assert isinstance(data.coords, expected_cls)



from astropy.coordinates import SkyCoord
from astropy.table import QTable
from glue.core import BaseData
from glue_jupyter.bqplot.image import BqplotImageView
from glue_jupyter.table import TableViewer

from jdaviz.core.events import (AddDataToViewerMessage,
                                RemoveDataFromViewerMessage,
                                TableClickMessage)
from jdaviz.core.astrowidgets_api import AstrowidgetsImageViewerMixin
from jdaviz.core.registries import viewer_registry
from jdaviz.core.freezable_state import FreezableBqplotImageViewerState
from jdaviz.configs.default.plugins.viewers import JdavizViewerMixin
from jdaviz.configs.specviz.plugins.viewers import Spectrum1DViewer, Spectrum2DViewer

__all__ = ['MosvizImageView', 'MosvizProfile2DView',
           'MosvizProfileView', 'MosvizTableViewer']


@viewer_registry("mosviz-image-viewer", label="Image 2D (Mosviz)")
class MosvizImageView(JdavizViewerMixin, BqplotImageView, AstrowidgetsImageViewerMixin):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom'],
                    ['jdaviz:panzoom'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = None
    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_astrowidgets_api()
        self._subscribe_to_layers_update()

        self.state.show_axes = False  # Axes are wrong anyway
        self.figure.fig_margin = {'left': 0, 'bottom': 0, 'top': 0, 'right': 0}

        self.data_menu._obj.dataset.add_filter('is_image_not_spectrum')
        self.data_menu._obj.dataset.add_filter('same_mosviz_row')

    def data(self, cls=None):
        return [layer_state.layer.get_object(cls=cls or self.default_class)
                for layer_state in self.state.layers
                if hasattr(layer_state, 'layer') and
                isinstance(layer_state.layer, BaseData)]

    # NOTE: This is currently only for debugging. It is not used in app.
    def _mark_targets(self):
        table_data = self.jdaviz_app.data_collection['MOS Table']

        if ("R.A." not in table_data.component_ids() or
                "Dec." not in table_data.component_ids()):
            return

        ra = table_data["R.A."]
        dec = table_data["Dec."]
        sky = SkyCoord(ra, dec, unit='deg')
        t = QTable({'coord': sky})
        self.add_markers(t, use_skycoord=True, marker_name='Targets')


@viewer_registry("mosviz-profile-2d-viewer", label="Spectrum 2D (Mosviz)")
class MosvizProfile2DView(Spectrum2DViewer):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['mosviz:homezoom'],
                    ['mosviz:boxzoom', 'mosviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['mosviz:panzoom', 'mosviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['bqplot:xrange'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    _state_cls = FreezableBqplotImageViewerState

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_menu._obj.dataset.add_filter('same_mosviz_row')


@viewer_registry("mosviz-profile-viewer", label="Profile 1D")
class MosvizProfileView(Spectrum1DViewer):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['mosviz:homezoom'],
                    ['mosviz:boxzoom', 'mosviz:xrangezoom', 'jdaviz:yrangezoom'],  # noqa
                    ['mosviz:panzoom', 'mosviz:panzoom_x', 'jdaviz:panzoom_y'],  # noqa
                    ['bqplot:xrange'],
                    ['jdaviz:selectline'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_menu._obj.dataset.add_filter('same_mosviz_row')


@viewer_registry("mosviz-table-viewer", label="Table (Mosviz)")
class MosvizTableViewer(TableViewer, JdavizViewerMixin):
    def __init__(self, session, *args, **kwargs):
        super().__init__(session, *args, **kwargs)

        self.figure_widget.observe(self._on_row_selected, names=['highlighted'])
        # enable scrolling: # https://github.com/glue-viz/glue-jupyter/pull/287
        self.widget_table.scrollable = True

        self._selected_data = {}
        self._shared_image = False
        self.row_selection_in_progress = False

        self._on_row_selected_begin = None
        self._on_row_selected_end = None

    @property
    def _default_table_viewer_reference_name(self):
        return self.jdaviz_helper._default_table_viewer_reference_name

    @property
    def _default_spectrum_viewer_reference_name(self):
        return self.jdaviz_helper._default_spectrum_viewer_reference_name

    @property
    def _default_spectrum_2d_viewer_reference_name(self):
        return self.jdaviz_helper._default_spectrum_2d_viewer_reference_name

    @property
    def _default_image_viewer_reference_name(self):
        return self.jdaviz_helper._default_image_viewer_reference_name

    def redraw(self):

        # Overload to hide components - we do this via overloading instead of
        # checking for changes in self.figure_widget.data because some components
        # might be added inplace to the dataset.

        if self.figure_widget.data is None:
            self.figure_widget.hidden_components = []
        else:
            components_str = [cid.label for cid in self.figure_widget.data.main_components]
            hidden = []
            for colname in ('Images', '1D Spectra', '2D Spectra'):
                if colname in components_str:
                    hidden.append(self.figure_widget.data.id[colname])
            self.figure_widget.hidden_components = hidden

        super().redraw()

    @property
    def nrows(self):
        return self.widget_table.get_state()['total_length']

    @property
    def current_row(self):
        return self.widget_table.highlighted

    def select_row(self, n):
        if n < 0 or n >= self.nrows:
            raise ValueError("n must be between 0 and {}".format(self.nrows-1))

        # compute and set the appropriate page
        # NOTE: traitlets won't detect internal changes to a dict
        options = self.widget_table.get_state()['options']
        page = int(n / options['itemsPerPage']) + 1
        if options['page'] != page:
            self.widget_table.set_state({'options': {**options, 'page': page}})
            self.widget_table.send_state()
        # select and highlight the row
        self.widget_table.highlighted = n

    def next_row(self):
        current_row = self.current_row
        new_row = 0 if current_row == self.nrows - 1 else current_row + 1
        self.select_row(new_row)

    def prev_row(self):
        current_row = self.current_row
        new_row = self.nrows - 1 if current_row == 0 else current_row - 1
        self.select_row(new_row)

    def _on_row_selected(self, event):
        if self._on_row_selected_begin:
            self._on_row_selected_begin(event)

        self.row_selection_in_progress = True
        # Grab the index of the latest selected row
        selected_index = event['new']
        mos_data = self.session.data_collection['MOS Table']

        # plugin data entries: select all in new row, deselect all others
        for data_item in self.jdaviz_app.data_collection:
            if data_item.meta.get('Plugin') is not None:
                if data_item.meta.get('mosviz_row') == selected_index:
                    self.session.hub.broadcast(AddDataToViewerMessage(
                        self._default_spectrum_viewer_reference_name, data_item.label, sender=self))
                else:
                    self.session.hub.broadcast(RemoveDataFromViewerMessage(
                        self._default_spectrum_viewer_reference_name, data_item.label, sender=self))

        for component in mos_data.components:
            comp_data = mos_data.get_component(component).data
            selected_data = comp_data[selected_index]

            if component.label == '1D Spectra':
                prev_data = self._selected_data.get(self._default_spectrum_viewer_reference_name)
                if prev_data != selected_data:
                    if prev_data:
                        # This covers the cases where data is unit converted
                        # and the name is modified
                        all_prev_data = [x
                                         for x in self.session.data_collection.labels
                                         if prev_data in x]
                        for modified_prev_data in all_prev_data:
                            if modified_prev_data:
                                remove_data_from_viewer_message = RemoveDataFromViewerMessage(
                                    self._default_spectrum_viewer_reference_name,
                                    modified_prev_data, sender=self
                                )
                                # reset the counter in the spectrum viewer's color cycler
                                # so that the newly selected row is displayed in gray and
                                # future additions will have other colors:
                                spectrum_viewer = self.jdaviz_app.get_viewer(
                                    self._default_spectrum_viewer_reference_name
                                )
                                spectrum_viewer.color_cycler.reset()

                                self.session.hub.broadcast(remove_data_from_viewer_message)

                    add_data_to_viewer_message = AddDataToViewerMessage(
                        self._default_spectrum_viewer_reference_name,
                        selected_data, sender=self
                    )
                    self.session.hub.broadcast(add_data_to_viewer_message)

                    self._selected_data[
                        self._default_spectrum_viewer_reference_name
                    ] = selected_data

            if component.label == '2D Spectra':
                prev_data = self._selected_data.get(self._default_spectrum_2d_viewer_reference_name)
                if prev_data != selected_data:
                    if prev_data:
                        remove_data_from_viewer_message = RemoveDataFromViewerMessage(
                            self._default_spectrum_2d_viewer_reference_name,
                            prev_data, sender=self
                        )
                        self.session.hub.broadcast(remove_data_from_viewer_message)

                    add_data_to_viewer_message = AddDataToViewerMessage(
                        self._default_spectrum_2d_viewer_reference_name,
                        selected_data, sender=self
                    )
                    self.session.hub.broadcast(add_data_to_viewer_message)

                    self._selected_data[
                        self._default_spectrum_2d_viewer_reference_name
                    ] = selected_data

            if component.label == 'Images':
                prev_data = self._selected_data.get(self._default_image_viewer_reference_name)
                if prev_data != selected_data:
                    if prev_data:
                        remove_data_from_viewer_message = RemoveDataFromViewerMessage(
                            self._default_image_viewer_reference_name, prev_data, sender=self)
                        self.session.hub.broadcast(remove_data_from_viewer_message)

                    add_data_to_viewer_message = AddDataToViewerMessage(
                        self._default_image_viewer_reference_name, selected_data, sender=self)
                    self.session.hub.broadcast(add_data_to_viewer_message)

                    self._selected_data[self._default_image_viewer_reference_name] = selected_data

        message = TableClickMessage(selected_index=selected_index,
                                    shared_image=self._shared_image,
                                    sender=self)
        self.session.hub.broadcast(message)

        self.row_selection_in_progress = False

        if self._on_row_selected_end:
            self._on_row_selected_end(event)

    def set_plot_axes(self, *args, **kwargs):
        return
